# coding: utf8

import clinica.engine as ce


class StatisticsSurfaceCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 'statistics-surface'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Surface-based mass-univariate analysis with SurfStat:\nhttp://clinica.run/doc/Pipelines/Stats_Surface/'  # noqa

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        from colorama import Fore
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group('%sMandatory arguments%s' % (Fore.BLUE , Fore.RESET))
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("subject_visits_with_covariates_tsv",
                                  help='TSV file containing a list of subjects with their sessions and all the covariates and factors needed for the GLM.')  # noqa
        clinica_comp.add_argument("design_matrix",
                                  help='String to define the design matrix that fits into the GLM, e.g. 1 + group + sex + age.')
        clinica_comp.add_argument("contrast",
                                  help='String to define the contrast matrix for the GLM, e.g. group. Please note that, when you want to perform negative correlation, the sign is ignored by the command line.')  # noqa
        clinica_comp.add_argument("string_format",
                                  help='String to define the format of the columns in the TSV file, e.g., %%s %%s %%s %%f if the columns contain a string, a string, a string and a number, respectively.')  # noqa
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        clinica_comp.add_argument("glm_type",
                                  help='String based on the GLM type for the hypothesis. You can choose between group_comparison and correlation.')  # noqa
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-fwhm", "--full_width_at_half_maximum",
                              type=int, default=20,
                              help='FWHM for the surface smoothing (default: --full_width_at_half_maximum 20).')
        optional.add_argument("-ft", "--feature_type",
                              type=str, default=None,
                              help='Type of surface-based feature: cortical_thickness or pet_fdg_projection (default: --feature_type cortical_thickness). Note: noddi_projection_ndi, noddi_projection_odi, noddi_projection_fiso, dti_projection_fa, dti_projection_md, dti_projection_rd or dti_projection_ad exists but are not yet officially released.')  # noqa
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results')
        clinica_opt.add_argument("-np", "--n_procs",
                                 metavar=('N'), type=int,
                                 help='Number of cores used to run in parallel')
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-cf", "--custom_file",
                              type=str, default=None,
                              help='Pattern of file inside CAPS directory using @subject, @session, @fwhm, @hemi. No --feature_type must be specified in order to use this flag. If you use this flag, you must specify a label with the --feature_label flag). See Wiki for an example.')  # noqa
        advanced.add_argument("-fl", "--feature_label",
                              type=str, default=None,
                              help='Name of the feature type, it will be saved on the CAPS _measure-FEATURE_LABEL key-value association.')  # noqa
        advanced.add_argument("-tup", "--threshold_uncorrected_pvalue",
                              type=float, default=0.001,
                              help='Threshold to display the uncorrected p-value (--threshold_uncorrected_pvalue 0.001).')  # noqa
        advanced.add_argument("-tcp", "--threshold_corrected_pvalue",
                              type=float, default=0.05,
                              help='Threshold to display the corrected p-value (default: --threshold_corrected_pvalue 0.05)')  # noqa
        advanced.add_argument("-ct", "--cluster_threshold",
                              type=float, default=0.001,
                              help='Threshold to define a cluster in the process of cluster-wise correction (default: --cluster_threshold 0.001).')  # noqa

    def run_command(self, args):
        """
        Run the pipelines with defined args
        """
        from clinica.pipelines.statistics_surface.statistics_surface_pipeline import StatisticsSurface
        from clinica.pipelines.statistics_surface.statistics_surface_utils import check_inputs
        from clinica.utils.stream import cprint
        import os

        if args.feature_type is not None:
            if args.custom_file is not None:
                raise Exception('--feature_type and --custom_file are mutually exclusive : you must choose between one or the other. See documentation for more informations.')
            if args.feature_label is not None:
                raise Exception('--feature_label should not be used with --feature_type')
            # FreeSurfer cortical thickness
            if args.feature_type == 'cortical_thickness':
                args.custom_file = '@subject/@session/t1/freesurfer_cross_sectional/@subject_@session/surf/@hemi.thickness.fwhm@fwhm.fsaverage.mgh'
                args.feature_label = 'ct'
            # PET cortical projection
            elif args.feature_type == 'pet_fdg_projection':
                args.custom_file = '@subject/@session/pet/surface/@subject_@session_task-rest_acq-fdg_pet_space-fsaverage_suvr-pons_pvc-iy_hemi-@hemi_fwhm-@fwhm_projection.mgh'
                args.feature_label = 'fdg'
            # NODDI, NDI, ODI and FISO
            elif args.feature_type == 'noddi_projection_ndi':
                args.custom_file = '@subject/@session/noddi/postprocessing/noddi-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-ficvf_hemi-@hemi.mgh'
                args.feature_label = 'NDI'
            elif args.feature_type == 'noddi_projection_fiso':
                args.custom_file = '@subject/@session/noddi/postprocessing/noddi-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-fiso_hemi-@hemi.mgh'
                args.feature_label = 'FISO'
            elif args.feature_type == 'noddi_projection_odi':
                args.custom_file = '@subject/@session/noddi/postprocessing/noddi-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-odi_hemi-@hemi.mgh'
                args.feature_label = 'ODI'
            # DTI fa, md, rd and ad
            elif args.feature_type == 'dti_projection_fa':
                args.custom_file = '@subject/@session/noddi/postprocessing/dti-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-fa_hemi-@hemi.mgh'
                args.feature_label = 'FA'
            elif args.feature_type == 'dti_projection_md':
                args.custom_file = '@subject/@session/noddi/postprocessing/dti-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-md_hemi-@hemi.mgh'
                args.feature_label = 'MD'
            elif args.feature_type == 'dti_projection_rd':
                args.custom_file = '@subject/@session/noddi/postprocessing/dti-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-rd_hemi-@hemi.mgh'
                args.feature_label = 'RD'
            elif args.feature_type == 'dti_projection_ad':
                args.custom_file = '@subject/@session/noddi/postprocessing/dti-register-vertex-fsaverage/cortex-projection/@subject_@session_OnFsaverage_fwhm-@fwhm_measure-ad_hemi-@hemi.mgh'
                args.feature_label = 'AD'
            else:
                raise Exception('Feature type ' + args.feature_type + ' not recognized. Use --custom_file to specify your own files (without --feature_type).')
        elif args.feature_type is None:
            if args.custom_file is None:
                cprint('No feature type selected : using cortical thickness as default value')
                args.custom_file = '@subject/@session/t1/freesurfer_cross_sectional/@subject_@session/surf/@hemi.thickness.fwhm@fwhm.fsaverage.mgh'
                args.feature_label = 'ct'
            else:
                cprint('Using custom features.')
                if args.feature_label is None:
                    raise Exception('You must specify a --feature_label when using the --custom_files flag')

        # Check if the group label has been existed, if yes, give an error to the users
        if os.path.exists(os.path.join(os.path.abspath(self.absolute_path(args.caps_directory)), 'groups', 'group-' + args.group_id)):
            error_message = 'group_id : ' + args.group_id + ' already exists, please choose another one or delete the existing folder and also the working directory and rerun the pipelines'
            raise Exception(error_message)

        pipeline = StatisticsSurface(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subject_visits_with_covariates_tsv))
        pipeline.parameters = {
            # pass these args by using self.parameters in a dictionary
            'design_matrix': args.design_matrix,
            'contrast': args.contrast,
            'str_format': args.string_format,
            'group_label': args.group_id,
            'glm_type': args.glm_type,
            'custom_file': args.custom_file,
            'feature_label': args.feature_label,
            'full_width_at_half_maximum': args.full_width_at_half_maximum,
            'threshold_uncorrected_pvalue': args.threshold_uncorrected_pvalue,
            'threshold_corrected_pvalue': args.threshold_corrected_pvalue,
            'cluster_threshold': args.cluster_threshold
        }
        pipeline.base_dir = self.absolute_path(args.working_directory)

        check_inputs(pipeline.caps_directory,
                     pipeline.parameters['custom_file'],
                     pipeline.parameters['full_width_at_half_maximum'],
                     pipeline.tsv_file)

        if args.n_procs:
            pipeline.run(plugin='MultiProc',
                         plugin_args={'n_procs': args.n_procs})
        else:
            print(pipeline.parameters)
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

import clinica.engine as ce


class T1VolumeExistingDartelCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 't1-volume-register-dartel'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Inter-subject registration using Dartel (using an existing Dartel template):\nhttp://clinica.run/doc/Pipelines/T1_Volume/'

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-s", "--smooth",
                              nargs='+', type=int, default=[8],
                              help="A list of integers specifying the different isomorphic FWHM in millimeters to smooth the image (default: --smooth 8).")
        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_OPTIONAL'])
        clinica_opt.add_argument("-tsv", "--subjects_sessions_tsv",
                                 help='TSV file containing a list of subjects with their sessions.')
        clinica_opt.add_argument("-wd", "--working_directory",
                                 help='Temporary directory to store pipelines intermediate results.')
        clinica_opt.add_argument("-np", "--n_procs",
                                 metavar=('N'), type=int,
                                 help='Number of cores used to run in parallel.')
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-t", "--tissues", nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to create flow fields to DARTEL template (default: GM, WM and CSF i.e. --tissues 1 2 3).')

    def run_command(self, args):
        """
        """
        from tempfile import mkdtemp
        from clinica.utils.stream import cprint
        from clinica.pipelines.t1_volume_existing_dartel.t1_volume_existing_dartel_pipeline import T1VolumeExistingDartel

        pipeline = T1VolumeExistingDartel(bids_directory=self.absolute_path(args.bids_directory),
                                          caps_directory=self.absolute_path(args.caps_directory),
                                          tsv_file=self.absolute_path(args.subjects_sessions_tsv),
                                          group_id=args.group_id)

        pipeline.parameters.update({
            'tissues': args.tissues
        })

        if args.working_directory is None:
            args.working_directory = mkdtemp()
        pipeline.base_dir = self.absolute_path(args.working_directory)

        if args.n_procs:
            pipeline.run(plugin='MultiProc', plugin_args={'n_procs': args.n_procs})
        else:
            pipeline.run()

        cprint("The " + self._name + " pipeline has completed. You can now delete the working directory (" + args.working_directory + ").")



# coding: utf8

"""
Utils for the implementation of t1-volume-tissue-segmentation pipeline
"""

import os
import numpy as np
from nipype.interfaces.spm.base import SPMCommandInputSpec, SPMCommand
from nipype.interfaces.base import TraitedSpec, OutputMultiPath, InputMultiPath, File, traits
from nipype.utils.filemanip import filename_to_list, list_to_filename

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def select_bids_images(subjects, sessions, image_type, bids_layout):
    """
    read the subjects_visits_tsv
    """
    if len(subjects) != len(sessions):
        raise RuntimeError("Subjects list and sessions list must have the same length.")

    return [select_image(subjects[i], sessions[i], image_type, bids_layout) for i in range(len(subjects))]


def select_image(participant_id, session_id, image_type, bids_layout):
    """
    starting from the tsv with the participant_id and the session_id look for the corresponding image in the bids directory
    """
    import warnings

    if participant_id.startswith('sub-'):
        participant_id = participant_id[4:]
    if session_id.startswith('ses-'):
        session_id = session_id[4:]

    selected_images = bids_layout.get(subject=participant_id, session=session_id, type=image_type,
                                      return_type='file', extensions='nii.gz')
    if len(selected_images) == 0:
        selected_images = bids_layout.get(subject=participant_id, session=session_id, type=image_type,
                                          return_type='file', extensions='nii')
    if len(selected_images) == 0:
        raise RuntimeError('No ' + image_type + ' images were found for participant ' + participant_id
                           + ' and session ' + session_id)
    if len(selected_images) > 1:
        warnings.warn('Several ' + image_type + ' images were found for participant ' + participant_id
                      + ' and session ' + session_id, RuntimeWarning)
    return selected_images[0]


def group_nested_images_by_subject(class_images, zip_files=False):
    """

    """
    from clinica.utils.io import zip_nii

    if zip_files:
        return [zip_nii([s for tissue in subject for s in tissue], True) for subject in class_images]

    return [[s for tissue in subject for s in tissue] for subject in class_images]


def get_tissue_tuples(tissue_map, tissue_classes, dartel_tissues, save_warped_unmodulated, save_warped_modulated):
    """
    Method to obtain the list of tuples, one for each tissue class, with the following fields:
         - tissue probability map (4D), 1-based index to frame
         - number of gaussians
         - which maps to save [Native, DARTEL] - a tuple of two boolean values
         - which maps to save [Unmodulated, Modulated] - a tuple of two boolean values

    :param tissue_map: Path to tissue maps
    :param tissue_classes: Classes of images to obtain from segmentation. Ex: [1,2,3] is GM, WM and CSF
    :param dartel_tissues: Classes of images to save for DARTEL template calculation. Ex: [1] is only GM'
    :param save_warped_unmodulated: Save warped unmodulated images for tissues specified in --tissue_classes
    :param save_warped_modulated: Save warped modulated images for tissues specified in --tissue_classes
    :return: List of tuples according to NewSegment input por tissues
    """
    tissues = []

    for i in range(1, 7):
        n_gaussians = 2

        if i == 4 or i == 5:
            n_gaussians = i - 1

        native_space = False
        dartel_input = False
        warped_unmodulated = False
        warped_modulated = False

        if i in tissue_classes:
            native_space = True
            if save_warped_unmodulated:
                warped_unmodulated = True
            if save_warped_modulated:
                warped_modulated = True

        if i in dartel_tissues:
            dartel_input = True

        tissues.append(((tissue_map, i),
                        n_gaussians,
                        (native_space, dartel_input),
                        (warped_unmodulated, warped_modulated)))
    return tissues


class ApplySegmentationDeformationInput(SPMCommandInputSpec):

    deformation_field = File(
        exists=True,  mandatory=True,
        field='comp{1}.def',
        desc='SPM Segmentation deformation file')
    in_files = InputMultiPath(
        File(exists=True), mandatory=True,
        field='out{1}.pull.fnames',
        desc='Files on which deformation field is applied')
    interpolation = traits.Range(
        low=0, high=7,
        field='out{1}.pull.interp',
        desc='degree of b-spline used for interpolation')
    mask = traits.Int(
        0,
        usedefault=True,
        field='out{1}.pull.mask',
        desc='image masking')
    fwhm = traits.List(
        traits.Float(0),
        field='out{1}.pull.fwhm',
        minlen=3, maxlen=3,
        desc='3-element list (opt)')


class ApplySegmentationDeformationOutput(TraitedSpec):
    out_files = OutputMultiPath(File(exists=True),
                                desc='Transformed files')


class ApplySegmentationDeformation(SPMCommand):
    """ Uses spm to apply a deformation field obtained from Segmentation routine to a given file

    Examples
    --------

    >>> import clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_utils as seg_utils
    >>> inv = seg_utils.ApplySegmentationDeformation()
    >>> inv.inputs.in_files = 'T1w.nii'
    >>> inv.inputs.deformation = 'y_T1w.nii'
    >>> inv.run() # doctest: +SKIP
    """

    input_spec = ApplySegmentationDeformationInput
    output_spec = ApplySegmentationDeformationOutput

    _jobtype = 'util'
    _jobname = 'defs'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'deformation_field':
            return np.array([list_to_filename(val)], dtype=object)
        if opt == 'in_files':
            return np.array(filename_to_list(val), dtype=object)
        return val

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_files'] = []
        for filename in self.inputs.in_files:
            _, fname = os.path.split(filename)
            outputs['out_files'].append(os.path.realpath('w%s' % fname))
        return outputs



# coding: utf8

"""
Redirect stream and log
"""
import sys

clinica_verbose = False


class FilterOut(object):
    def __init__(self, stdout):
        self.stdout = stdout

    def write(self, text):
        import re
        if not text:
            return
        if re.match('^(@clinica@)', text):
            self.stdout.write(text.replace("@clinica@", ""))
            self.stdout.flush()

    def flush(self): self.stdout.flush()

    def __enter__(self):
        self.origin_stdout = sys.stdout
        self.flush()

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.origin_stdout
        self.flush()


def active_cprint():
    sys.stdout = FilterOut(sys.stdout)


def cprint(msg):
    global clinica_verbose
    if clinica_verbose is True:
        print(msg)
    else:
        print("@clinica@ %s" % msg)
    sys.stdout.flush()



# coding: utf8

import clinica.pipelines.engine as cpe

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class T1VolumeExistingDartel(cpe.Pipeline):
    """T1VolumeExistingDartel - Reuse existing Dartel template.

    Args:
        input_dir: A BIDS directory.
        output_dir: An empty output directory where CAPS structured data will
        be written.  subjects_sessions_list: The Subjects-Sessions list file
        (in .tsv format).

    Returns:
        A clinica pipeline object containing the T1VolumeExistingDartel pipeline.

    Raises:


    """

    def __init__(self, bids_directory=None, caps_directory=None, tsv_file=None, name=None, group_id='default'):
        super(T1VolumeExistingDartel, self).__init__(bids_directory, caps_directory, tsv_file, name)

        if not group_id.isalnum():
            raise ValueError('Not valid group_id value. It must be composed only by letters and/or numbers')

        self._group_id = group_id

        # Default parameters
        self._parameters = {'tissues': [1, 2, 3],
                            'iteration_parameters': None,
                            'optimization_parameters': None,
                            'regularization_form': None
                            }

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['dartel_input_images', 'dartel_iteration_templates']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return ['dartel_flow_fields']

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """

        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio

        tissue_names = {1: 'graymatter',
                        2: 'whitematter',
                        3: 'csf',
                        4: 'bone',
                        5: 'softtissue',
                        6: 'background'
                        }

        # Dartel Input Tissues DataGrabber
        # =================================
        dartel_input_reader = npe.MapNode(nio.DataGrabber(infields=['subject_id', 'session',
                                                                    'subject_repeat', 'session_repeat',
                                                                    'tissue'],
                                                          outfields=['out_files']),
                                          name="dartel_input_reader",
                                          iterfield=['tissue'])

        dartel_input_reader.inputs.base_directory = self.caps_directory
        dartel_input_reader.inputs.template = 'subjects/%s/%s/t1/spm/segmentation/dartel_input/%s_%s_T1w_segm-%s_dartelinput.nii*'
        dartel_input_reader.inputs.subject_id = self.subjects
        dartel_input_reader.inputs.session = self.sessions
        dartel_input_reader.inputs.tissue = [tissue_names[t] for t in self.parameters['tissues']]
        dartel_input_reader.inputs.subject_repeat = self.subjects
        dartel_input_reader.inputs.session_repeat = self.sessions
        dartel_input_reader.inputs.sort_filelist = False

        # Dartel Templates DataGrabber
        # ============================
        templates_reader = npe.MapNode(nio.DataGrabber(infields=['iteration'],
                                                       outfields=['out_files']),
                                       name="templates_reader",
                                       iterfield=['iteration'])
        templates_reader.inputs.base_directory = self.caps_directory
        templates_reader.inputs.template = 'groups/group-' + self._group_id + '/t1/group-' + \
                                           self._group_id + '_iteration-%d_template.nii*'
        templates_reader.inputs.iteration = range(1, 7)
        templates_reader.inputs.sort_filelist = False

        self.connect([
            (dartel_input_reader, self.input_node, [('out_files', 'dartel_input_images')]),
            (templates_reader, self.input_node, [('out_files', 'dartel_iteration_templates')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        import os.path as op
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        import re
        from clinica.utils.io import zip_nii

        # Writing flowfields into CAPS
        # ============================
        write_flowfields_node = npe.MapNode(name='write_flowfields_node',
                                            iterfield=['container', 'flow_fields'],
                                            interface=nio.DataSink(infields=['flow_fields']))
        write_flowfields_node.inputs.base_directory = self.caps_directory
        write_flowfields_node.inputs.parameterization = False
        write_flowfields_node.inputs.container = ['subjects/' + self.subjects[i] + '/' + self.sessions[i] +
                                                  '/t1/spm/dartel/group-' + self._group_id
                                                  for i in range(len(self.subjects))]
        write_flowfields_node.inputs.regexp_substitutions = [
            (r'(.*)_Template(\.nii(\.gz)?)$', r'\1\2'),
            (r'(.*)c1(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-graymatter\3'),
            (r'(.*)c2(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-whitematter\3'),
            (r'(.*)c3(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-csf\3'),
            (r'(.*)c4(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-bone\3'),
            (r'(.*)c5(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-softtissue\3'),
            (r'(.*)c6(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-background\3'),
            (r'(.*)r(sub-.*)(\.nii(\.gz)?)$', r'\1\2\3'),
            (r'(.*)_dartelinput(\.nii(\.gz)?)$', r'\1\2'),
            (r'(.*)flow_fields/u_(sub-.*)_segm-.*(\.nii(\.gz)?)$',
             r'\1\2_target-' + re.escape(self._group_id) + r'_transformation-forward_deformation\3'),
            (r'trait_added', r'')
        ]

        self.connect([
            (self.output_node, write_flowfields_node, [(('dartel_flow_fields', zip_nii, True), 'flow_fields')])
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.io import unzip_nii
        import clinica.pipelines.t1_volume_existing_dartel.t1_volume_existing_dartel_utils as utils

        # Unzipping
        # =========
        unzip_dartel_input_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                             output_names=['out_file'],
                                                             function=unzip_nii),
                                              name='unzip_dartel_input_node',
                                              iterfield=['in_file'])
        unzip_templates_node = npe.Node(nutil.Function(input_names=['in_file'],
                                                       output_names=['out_file'],
                                                       function=unzip_nii),
                                        name='unzip_templates_node')
        # DARTEL with existing template
        # =============================
        dartel_existing_template = npe.MapNode(utils.DARTELExistingTemplate(),
                                               name='dartel_existing_template',
                                               iterfield=['image_files'])
        if self.parameters['optimization_parameters'] is not None:
            dartel_existing_template.inputs.optimization_parameters = self.parameters['optimization_parameters']
        if self.parameters['regularization_form'] is not None:
            dartel_existing_template.inputs.regularization_form = self.parameters['regularization_form']

        # Connection
        # ==========
        self.connect([
            (self.input_node, unzip_dartel_input_node, [('dartel_input_images', 'in_file')]),
            (self.input_node, unzip_templates_node, [('dartel_iteration_templates', 'in_file')]),
            (unzip_dartel_input_node, dartel_existing_template, [(('out_file', utils.prepare_dartel_input_images),
                                                                  'image_files')]),
            (unzip_templates_node, dartel_existing_template, [(('out_file', utils.create_iteration_parameters,
                                                                self.parameters['iteration_parameters']),
                                                               'iteration_parameters')]),
            (dartel_existing_template, self.output_node, [('dartel_flow_fields', 'dartel_flow_fields')])
        ])



# coding: utf-8

from clinica.iotools.abstract_converter import Converter

__author__ = "Jorge Samper-Gonzalez and Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = [""]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class AdniToBids(Converter):

    def get_modalities_supported(self):
        """
        Return a list of modalities supported

        Returns: a list containing the modalities supported by the converter (T1, PET_FDG, PET_AV45)

        """
        return ['T1', 'PET_FDG', 'PET_AV45', 'DWI', 'fMRI', 'FLAIR']

    def check_adni_dependencies(self):
        """
        Check the dependencies of ADNI converter
        """
        from clinica.utils.check_dependency import is_binary_present

        list_binaries = ['dcm2nii', 'dcm2niix']

        for binary in list_binaries:
            if not is_binary_present(binary):
                raise RuntimeError(
                    '%s is not present '
                    'in your PATH environment.' % binary)

    def convert_clinical_data(self,  clinical_data_dir, out_path):
        """
        Convert the clinical data of ADNI specified into the file clinical_specifications_adni.xlsx

        Args:
            clinical_data_dir:  path to the clinical data directory
            out_path: path to the BIDS directory

        """
        from os import path
        import os
        import pandas as pd
        import clinica.iotools.bids_utils as bids
        from clinica.utils.stream import cprint
        import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils

        clinic_specs_path = path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'data',
                                      'clinical_specifications_adni.xlsx')
        try:
            os.path.exists(out_path)
        except IOError:
            print('BIDS folder not found.')
            raise

        bids_ids = bids.get_bids_subjs_list(out_path)
        bids_subjs_paths = bids.get_bids_subjs_paths(out_path)

        if not bids_ids:
            adni_merge_path = path.join(clinical_data_dir, 'ADNIMERGE.csv')
            adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
            bids_ids = ['sub-ADNI' + subj.replace('_', '') for subj in list(adni_merge.PTID.unique())]
            bids_subjs_paths = [path.join(out_path, subj) for subj in bids_ids]

        # -- Creation of participant.tsv --
        cprint("Creating participants.tsv...")
        participants_df = bids.create_participants_df('ADNI', clinic_specs_path, clinical_data_dir, bids_ids)

        # Replace the original values with the standard defined by the AramisTeam
        participants_df['sex'] = participants_df['sex'].replace('Male', 'M')
        participants_df['sex'] = participants_df['sex'].replace('Female', 'F')

        participants_df.to_csv(path.join(out_path, 'participants.tsv'), sep='\t', index=False, encoding='utf-8')

        # -- Creation of sessions.tsv --
        cprint("Creating sessions files...")
        adni_utils.create_adni_sessions_dict(bids_ids, clinic_specs_path, clinical_data_dir, bids_subjs_paths)

        # -- Creation of scans files --
        cprint('Creating scans files...')
        adni_utils.create_adni_scans_files(clinic_specs_path, bids_subjs_paths, bids_ids)

    def convert_images(self, source_dir, clinical_dir, dest_dir, subjs_list_path='', modalities=['T1', 'PET_FDG', 'PET_AV45', 'DWI']):
        """
        Convert the images of ADNI

        Args:
            source_dir: path to the ADNI directory
            clinical_dir: path to the clinical data directory
            dest_dir: path to the BIDS directory
            subjs_list_path: list of subjects to process
            modalities: modalities to convert (T1, PET_FDG, PET_AV45, DWI, FLAIR)

        """

        import os
        from os import path
        import pandas as pd
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_t1 as adni_t1
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_av45_pet as adni_av45
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_fdg_pet as adni_fdg
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_dwi as adni_dwi
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_fmri as adni_fmri
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_flair as adni_flair
        from clinica.utils.stream import cprint

        adni_merge_path = path.join(clinical_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path)

        # Load a file with subjects list or compute all the subjects
        if subjs_list_path is not None and subjs_list_path != '':
            cprint('Loading a subjects lists provided by the user...')
            subjs_list = [line.rstrip('\n') for line in open(subjs_list_path)]
        else:
            cprint('Using all the subjects contained into the ADNIMERGE.csv file...')
            subjs_list = list(adni_merge['PTID'].unique())

        # Create the output folder if is not already existing
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
            os.makedirs(path.join(dest_dir, 'conversion_info'))

        if 'T1' in modalities:
            adni_t1.convert_adni_t1(source_dir, clinical_dir, dest_dir, subjs_list)

        if 'PET_FDG' in modalities:
            adni_fdg.convert_adni_fdg_pet(source_dir, clinical_dir, dest_dir, subjs_list)

        if 'PET_AV45' in modalities:
            adni_av45.convert_adni_av45_pet(source_dir, clinical_dir, dest_dir, subjs_list)

        if 'DWI' in modalities:
            adni_dwi.convert_adni_dwi(source_dir, clinical_dir, dest_dir, subjs_list)

        if 'fMRI' in modalities:
            adni_fmri.convert_adni_fmri(source_dir, clinical_dir, dest_dir, subjs_list)

        if 'FLAIR' in modalities:
            adni_flair.convert_adni_flair(source_dir, clinical_dir, dest_dir, subjs_list)



# coding: utf8

# WARNING: Don't put any import statement here except if it's absolutely
# necessary. Put it *inside* the different methods.
# Otherwise it will slow down the dynamic loading of the pipelines list by the
# command line tool.
import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
from nipype import config
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class DwiConnectome(cpe.Pipeline):
    """Connectome-based processing of DWI datasets.

    Todos:
        [ ] - Add registration of parcellation onto DWI space

    Args:
        input_dir: A BIDS directory.
        output_dir: An empty output directory where CAPS structured data will
        be written.
        subjects_sessions_list: The Subjects-Sessions list file (in .tsv
        format).

    Returns:
        A clinica pipeline object containing the DwiConnectome pipeline.

    Raises:

    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['t1_brain_file', 'wm_mask_file', 'dwi_file',
                'dwi_brainmask_file', 'grad_fsl', 'atlas_files']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return ['response', 'fod', 'tracts', 'nodes',
                'connectomes']

    def build_input_node(self):
        """Build and connect an input node to the pipeline.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.iotools.grabcaps import CAPSLayout
        from clinica.utils.stream import cprint

        from clinica.utils.exceptions import ClinicaCAPSError

        # Reading BIDS files
        # ==================

        # cprint('Loading CAPS folder...')
        caps_layout = CAPSLayout(self.caps_directory)
        # cprint('CAPS folder loaded')

        wm_mask_file = []
        dwi_file_space = []
        t1_brain_file = []
        dwi_file = []
        dwi_brainmask_file = []
        bvec = []
        bval = []
        grad_fsl = []
        aparc_aseg = []
        aparc_aseg_a2009s = []
        atlas_files = []
        cprint('Extracting files...')
        for i in range(len(self.subjects)):
            cprint('\t...subject \'' + str(
                    self.subjects[i][4:]) + '\', session \'' + str(
                    self.sessions[i][4:]) + '\'')

            wm_mask_file.append(caps_layout.get(freesurfer_file='wm.seg.mgz',
                                                session=self.sessions[i][4:],
                                                subject=self.subjects[i][4:],
                                                return_type='file')[0])

            dwi_file_space.append(
                    caps_layout.get(type='dwi', suffix='preproc',
                                    target='space',
                                    session=self.sessions[i][4:],
                                    extensions='nii.gz',
                                    subject=self.subjects[i][4:],
                                    return_type='id')[0])

            if dwi_file_space[i] == 'b0':
                t1_brain_file.append(
                        caps_layout.get(freesurfer_file='brain.mgz',
                                        session=self.sessions[i][4:],
                                        subject=self.subjects[i][4:],
                                        return_type='file')[0])

            aparc_aseg.append(
                    caps_layout.get(freesurfer_file='aparc\+aseg.mgz',
                                    session=self.sessions[i][4:],
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            aparc_aseg_a2009s.append(
                    caps_layout.get(freesurfer_file='aparc.a2009s\+aseg.mgz',
                                    session=self.sessions[i][4:],
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            atlas_files.append([aparc_aseg[i], aparc_aseg_a2009s[i]])

            dwi_file.append(
                    caps_layout.get(type='dwi', suffix='preproc',
                                    space=dwi_file_space[i],
                                    session=self.sessions[i][4:],
                                    extensions='nii.gz',
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            dwi_brainmask_file.append(
                    caps_layout.get(type='dwi', suffix='brainmask',
                                    space=dwi_file_space[i],
                                    session=self.sessions[i][4:],
                                    extensions='nii.gz',
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            bvec.append(
                    caps_layout.get(type='dwi', suffix='preproc',
                                    space=dwi_file_space[i],
                                    session=self.sessions[i][4:],
                                    extensions='bvec',
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            bval.append(
                    caps_layout.get(type='dwi', suffix='preproc',
                                    space=dwi_file_space[i],
                                    session=self.sessions[i][4:],
                                    extensions='bval',
                                    subject=self.subjects[i][4:],
                                    return_type='file')[0])

            grad_fsl.append((bvec[i], bval[i]))

        # Return an error if all the DWI files are not in the same space
        if any(a != dwi_file_space[0] for a in dwi_file_space):
            raise ClinicaCAPSError('Preprocessed DWI files are not all in the '
                                   'same space. Please process them separately '
                                   'using the appropriate subjects/sessions '
                                   '`.tsv` file (-tsv option).')

        elif dwi_file_space[0] == 'b0':
            self.parameters['dwi_space'] = 'b0'
            read_node = npe.Node(name="ReadingFiles",
                                 iterables=[
                                     ('wm_mask_file', wm_mask_file),
                                     ('t1_brain_file', t1_brain_file),
                                     ('dwi_file', dwi_file),
                                     ('dwi_brainmask_file', dwi_brainmask_file),
                                     ('grad_fsl', grad_fsl),
                                     ('atlas_files', atlas_files)
                                 ],
                                 synchronize=True,
                                 interface=nutil.IdentityInterface(
                                         fields=self.get_input_fields()))
            self.connect([
                (read_node, self.input_node, [('t1_brain_file', 't1_brain_file')]),
                (read_node, self.input_node, [('wm_mask_file', 'wm_mask_file')]),
                (read_node, self.input_node, [('dwi_file', 'dwi_file')]),
                (read_node, self.input_node, [('dwi_brainmask_file', 'dwi_brainmask_file')]),
                (read_node, self.input_node, [('grad_fsl', 'grad_fsl')]),
                (read_node, self.input_node, [('atlas_files', 'atlas_files')]),
            ])

        elif dwi_file_space[0] == 'T1w':
            self.parameters['dwi_space'] = 'T1w'
            read_node = npe.Node(name="ReadingFiles",
                                 iterables=[
                                     ('wm_mask_file', wm_mask_file),
                                     ('dwi_file', dwi_file),
                                     ('dwi_brainmask_file', dwi_brainmask_file),
                                     ('grad_fsl', grad_fsl),
                                     ('atlas_files', atlas_files)
                                 ],
                                 synchronize=True,
                                 interface=nutil.IdentityInterface(
                                         fields=self.get_input_fields()))
            self.connect([
                (read_node, self.input_node, [('wm_mask_file', 'wm_mask_file')]),
                (read_node, self.input_node, [('dwi_file', 'dwi_file')]),
                (read_node, self.input_node, [('dwi_brainmask_file', 'dwi_brainmask_file')]),
                (read_node, self.input_node, [('grad_fsl', 'grad_fsl')]),
                (read_node, self.input_node, [('atlas_files', 'atlas_files')]),
            ])

        else:
            raise ClinicaCAPSError('Bad preprocessed DWI space. Please check '
                                   'your CAPS folder.')

    def build_output_node(self):
        """Build and connect an output node to the pipeline.
        """

        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        import nipype.interfaces.io as nio
        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils

        # Writing CAPS
        # ============

        join_node = npe.JoinNode(name='JoinOutputs',
                                 joinsource='ReadingFiles',
                                 interface=nutil.IdentityInterface(
                                         fields=self.get_output_fields()))

        write_node = npe.MapNode(name='WritingCAPS',
                                 iterfield=['container'] +
                                           ['connectome_based_processing.@' + o
                                            for o in self.get_output_fields()],
                                 interface=nio.DataSink(
                                         infields=['connectome_based_processing.@' + o for o in
                                                   self.get_output_fields()]))
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.container = utils.get_containers(self.subjects,
                                                           self.sessions)
        write_node.inputs.substitutions = [('trait_added', '')]
        write_node.inputs.parameterization = False

        self.connect([
            # Writing CAPS
            (self.output_node, join_node, [('response', 'response')]),
            (self.output_node, join_node, [('fod', 'fod')]),
            (self.output_node, join_node, [('tracts', 'tracts')]),
            (self.output_node, join_node, [('nodes', 'nodes')]),
            (self.output_node, join_node, [('connectomes', 'connectomes')]),
            (join_node, write_node, [('response', 'connectome_based_processing.@response')]),
            (join_node, write_node, [('fod', 'connectome_based_processing.@fod')]),
            (join_node, write_node, [('tracts', 'connectome_based_processing.@tracts')]),
            (join_node, write_node, [('nodes', 'connectome_based_processing.@nodes')]),
            (join_node, write_node, [('connectomes', 'connectome_based_processing.@connectomes')]),
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.

        Notes:
            - If `FSLOUTPUTTYPE` environment variable is not set, `nipype` takes
            NIFTI by default.

        Todos:
            - [x] Detect space automatically.
            - [ ] Allow for custom parcellations (See TODOs in utils).

        """
        import nipype.interfaces.utility as niu
        import nipype.pipeline.engine as npe
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.freesurfer as fs
        import nipype.interfaces.mrtrix3 as mrtrix3
        from clinica.lib.nipype.interfaces.mrtrix3.reconst import EstimateFOD
        from clinica.lib.nipype.interfaces.mrtrix3.tracking import Tractography
        from clinica.utils.exceptions import ClinicaCAPSError
        from clinica.utils.stream import cprint
        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils

        # cprint('Building the pipeline...')

        # Nodes
        # =====

        # B0 Extraction
        # -------------
        split_node = npe.Node(name="B0Extraction",
                              interface=fsl.Split())
        split_node.inputs.output_type = "NIFTI_GZ"
        split_node.inputs.dimension = 't'
        select_node = npe.Node(name="B0Selection", interface=niu.Select())
        select_node.inputs.index = 0

        # B0 Brain Extraction
        # -------------------
        mask_node = npe.Node(name="BrainMasking",
                             interface=fsl.ApplyMask())
        mask_node.inputs.output_type = "NIFTI_GZ"

        # T1-to-B0 Registration
        # ---------------------
        t12b0_reg_node = npe.Node(name="T12B0Registration",
                                  interface=fsl.FLIRT())
        t12b0_reg_node.inputs.output_type = "NIFTI_GZ"

        # MGZ File Convertion
        # -------------------
        t1_brain_conv_node = npe.Node(name="T1BrainConvertion",
                                      interface=fs.MRIConvert())
        wm_mask_conv_node = npe.Node(name="WMMaskConvertion",
                                     interface=fs.MRIConvert())

        # WM Transformation
        # -----------------
        wm_transform_node = npe.Node(name="WMTransformation",
                                     interface=fsl.ApplyXFM())
        wm_transform_node.inputs.apply_xfm = True

        # Response Estimation
        # -------------------
        resp_estim_node = npe.Node(name="ResponseEstimation",
                                   interface=mrtrix3.ResponseSD())
        resp_estim_node.inputs.algorithm = 'tournier'

        # FOD Estimation
        # --------------
        fod_estim_node = npe.Node(name="FODEstimation",
                                  interface=EstimateFOD())
        fod_estim_node.inputs.algorithm = 'csd'

        # Tracts Generation
        # -----------------
        tck_gen_node = npe.Node(name="TractsGeneration",
                                interface=utils.Tractography())
        tck_gen_node.inputs.n_tracks = self.parameters['n_tracks']
        # BUG: Info package does not exist
        # from nipype.interfaces.mrtrix3.base import Info
        # from distutils.version import LooseVersion
        #
        # if Info.looseversion() >= LooseVersion("3.0"):
        #     tck_gen_node.inputs.select = self.parameters['n_tracks']
        # elif Info.looseversion() <= LooseVersion("0.4"):
        #     tck_gen_node.inputs.n_tracks = self.parameters['n_tracks']
        # else:
        #     from clinica.utils.exceptions import ClinicaException
        #     raise ClinicaException("Your MRtrix version is not supported.")

        # Nodes Generation
        # ----------------
        label_convert_node = npe.MapNode(name="LabelsConversion",
                                         iterfield=['in_file', 'in_config',
                                                    'in_lut', 'out_file'],
                                         interface=mrtrix3.LabelConvert())
        label_convert_node.inputs.in_config = utils.get_conversion_luts()
        label_convert_node.inputs.in_lut = utils.get_luts()

        # Connectome Generation
        # ---------------------
        # only the parcellation and output filename should be iterable, the tck
        # file stays the same.
        conn_gen_node = npe.MapNode(name="ConnectomeGeneration",
                                    iterfield=['in_parc', 'out_file'],
                                    interface=mrtrix3.BuildConnectome())

        # Print begin message
        # -------------------
        print_begin_message = npe.Node(
            interface=niu.Function(
                input_names=['in_bids_or_caps_file'],
                function=utils.print_begin_pipeline),
            iterfield='in_bids_or_caps_file',
            name='WriteBeginMessage')

        # Print end message
        # -----------------
        print_end_message = npe.Node(
            interface=niu.Function(
                input_names=['in_bids_or_caps_file', 'final_file', 'final_file'],
                function=utils.print_end_pipeline),
            iterfield=['in_bids_or_caps_file', 'final_file'],
            name='WriteEndMessage')

        # CAPS FIlenames Generation
        # -------------------------
        caps_filenames_node = npe.Node(name='CAPSFilenamesGeneration',
                                       interface=niu.Function(
                                               input_names='dwi_file',
                                               output_names=self.get_output_fields(),
                                               function=utils.get_caps_filenames))

        # Connections
        # ===========
        self.connect([
            (self.input_node, print_begin_message, [('dwi_file', 'in_bids_or_caps_file')]),  # noqa
        ])
        # WM in B0-space Transformation
        # -----------------------------
        if self.parameters['dwi_space'] == 'b0':
            self.connect([
                # MGZ Files Convertion
                (self.input_node, t1_brain_conv_node,
                 [('t1_brain_file', 'in_file')]),

                # noqa
                (self.input_node, wm_mask_conv_node,
                 [('wm_mask_file', 'in_file')]),
                # noqa
                # B0 Extraction
                (self.input_node, split_node, [('dwi_file', 'in_file')]),
                # noqa
                (split_node, select_node, [('out_files', 'inlist')]),
                # noqa
                # Masking
                (select_node, mask_node, [('out', 'in_file')]),  # B0 #noqa
                (self.input_node, mask_node,
                 [('dwi_brainmask_file', 'mask_file')]),
                # Brain mask #noqa #noqa
                # T1-to-B0 Registration
                (t1_brain_conv_node, t12b0_reg_node, [('out_file', 'in_file')]),
                # Brain #noqa
                (mask_node, t12b0_reg_node, [('out_file', 'reference')]),
                # B0 brain-masked #noqa
                # WM Transformation
                (wm_mask_conv_node, wm_transform_node,
                 [('out_file', 'in_file')]),  # Brain mask #noqa
                (mask_node, wm_transform_node, [('out_file', 'reference')]),
                # BO brain-masked #noqa
                (t12b0_reg_node, wm_transform_node,
                 [('out_matrix_file', 'in_matrix_file')]),
                # T1-to-B0 matrix file #noqa
            ])

        # Tractography
        # ------------
        self.connect([
            (self.input_node, caps_filenames_node, [('dwi_file', 'dwi_file')]),
            # Response Estimation
            (self.input_node, resp_estim_node, [('dwi_file', 'in_file')]),
            # Preproc. DWI #noqa
            (self.input_node, resp_estim_node,
             [('dwi_brainmask_file', 'in_mask')]),  # B0 brain mask #noqa
            (self.input_node, resp_estim_node, [('grad_fsl', 'grad_fsl')]),
            # bvecs and bvals #noqa
            (caps_filenames_node, resp_estim_node, [('response', 'wm_file')]),
            # output response filename #noqa
            # FOD Estimation
            (self.input_node, fod_estim_node, [('dwi_file', 'in_file')]),
            # Preproc. DWI #noqa
            (resp_estim_node, fod_estim_node, [('wm_file', 'wm_txt')]),
            # Response (txt file) #noqa
            (self.input_node, fod_estim_node,
             [('dwi_brainmask_file', 'mask_file')]),  # B0 brain mask #noqa
            (self.input_node, fod_estim_node, [('grad_fsl', 'grad_fsl')]),
            # T1-to-B0 matrix file #noqa
            (caps_filenames_node, fod_estim_node, [('fod', 'wm_odf')]),
            # output odf filename #noqa
            # Tracts Generation
            (fod_estim_node, tck_gen_node, [('wm_odf', 'in_file')]),
            # ODF file #noqa
            (caps_filenames_node, tck_gen_node, [('tracts', 'out_file')]),
            # output tck filename #noqa
            # Label Conversion
            (self.input_node, label_convert_node, [('atlas_files', 'in_file')]),
            # atlas image files #noqa
            (caps_filenames_node, label_convert_node, [('nodes', 'out_file')]),
            # converted atlas image filenames #noqa
            # Connectomes Generation
            (tck_gen_node, conn_gen_node, [('out_file', 'in_file')]),
            # output odf filename #noqa
            (label_convert_node, conn_gen_node, [('out_file', 'in_parc')]),
            # output odf filename #noqa
            (caps_filenames_node, conn_gen_node, [('connectomes', 'out_file')]),
            # output odf filename #noqa
        ])

        if self.parameters['dwi_space'] == 'b0':
            self.connect([
                (wm_transform_node, tck_gen_node, [('out_file', 'seed_image')])
                # ODF file #noqa
            ])
        elif self.parameters['dwi_space'] == 'T1w':
            self.connect([
                (
                    self.input_node, tck_gen_node,
                    [('wm_mask_file', 'seed_image')])
                # ODF file #noqa
            ])
        else:
            raise ClinicaCAPSError(
                    'Bad preprocessed DWI space. Please check your CAPS '
                    'folder.')

        # Outputs
        # ------
        self.connect([
            (resp_estim_node, self.output_node, [('wm_file', 'response')]),
            # T1-to-B0 matrix file #noqa
            (fod_estim_node, self.output_node, [('wm_odf', 'fod')]),
            # T1-to-B0 matrix file #noqa
            (tck_gen_node, self.output_node, [('out_file', 'tracts')]),
            # T1-to-B0 matrix file #noqa
            (label_convert_node, self.output_node, [('out_file', 'nodes')]),
            # T1-to-B0 matrix file #noqa
            (conn_gen_node, self.output_node, [('out_file', 'connectomes')]),
            # T1-to-B0 matrix file #noqa
            (self.input_node, print_end_message, [('dwi_file', 'in_bids_or_caps_file')]),  # noqa
            (conn_gen_node,   print_end_message, [('out_file', 'final_file')]),  # noqa
        ])

        # cprint('Pipeline built')



# coding: utf-8

"""
 Module for converting T1 of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_t1(source_dir, csv_dir, dest_dir, subjs_list=None):
    """

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination directory
        subjs_list: subjects list

    Returns:

    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
        subjs_list = list(adni_merge.PTID.unique())
    # new_download: indicates the version of csv Downloaded
    if path.isfile(path.join(csv_dir, 'IDA_MR_Metadata_Listing.csv')):
        new_download = False
    else:
        new_download = True
    if new_download is True:
        if '127_S_5200' in subjs_list:
            subjs_list.remove('127_S_5200')
        if '027_S_5083' in subjs_list:
            subjs_list.remove('027_S_5083')
        if '941_S_4365' in subjs_list:
            subjs_list.remove('941_S_4365')
        if '037_S_4028' in subjs_list:
            subjs_list.remove('037_S_4028')

    cprint('Calculating paths of T1 images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list, new_download)
    cprint('Paths of T1 images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 't1')
    cprint(Fore.GREEN + 'T1 conversion done.' + Fore.RESET)


def compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list, new_download):
    """

    Compute paths to t1 images of ADNI.

    :param source_dir:
    :param csv_dir:
    :param dest_dir:
    :param subjs_list:
    :return: a pandas dataframe
    """

    import pandas as pd
    import operator
    from os import path, walk, mkdir
    from functools import reduce

    t1_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                 'Study_ID', 'Field_Strength', 'Series_ID', 'Image_ID', 'Original']

    t1_df = pd.DataFrame(columns=t1_col_df)
    adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
    # adni_screening_path = path.join(clinical_dir, 'ADNI_ScreeningList_8_22_12.csv')
    if new_download is False:
        ida_meta_path = path.join(csv_dir, 'IDA_MR_Metadata_Listing.csv')
        ida_meta = pd.io.parsers.read_csv(ida_meta_path, sep=',', low_memory=False)

    mprage_meta_path = path.join(csv_dir, 'MPRAGEMETA.csv')
    mri_quality_path = path.join(csv_dir, 'MRIQUALITY.csv')
    mayo_mri_qc_path = path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv')

    adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',', low_memory=False)
    mprage_meta = pd.io.parsers.read_csv(mprage_meta_path, sep=',', low_memory=False)
    mri_quality = pd.io.parsers.read_csv(mri_quality_path, sep=',', low_memory=False)
    mayo_mri_qc = pd.io.parsers.read_csv(mayo_mri_qc_path, sep=',', low_memory=False)
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'T1']

    for subj in subjs_list:
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        # Sort the values by examination date
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mprage_meta_subj = mprage_meta[mprage_meta.SubjectID == subj]
        mprage_meta_subj = mprage_meta_subj.sort_values('ScanDate')

        mri_quality_subj = mri_quality[mri_quality.RID == int(subj[-4:])]
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']
        visits = visits_to_timepoints_t1(subj, mprage_meta_subj, adnimerge_subj)

        keys = list(visits.keys())
        keys.sort()
        for visit_info in visits.keys():
            if visit_info[1] == 'ADNI1':
                if not new_download:
                    ida_meta_subj = ida_meta[ida_meta.Subject == subj]
                    image_dict = adni1_image(
                            subj, visit_info[0], visits[visit_info],
                            mprage_meta_subj, ida_meta_subj,
                            mri_quality_subj, mayo_mri_qc_subj)
                else:
                    image_dict = adni1_image_refactoring(csv_dir, adni_merge, subj, visit_info[0], visits[visit_info],
                                                         mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj)
            elif visit_info[1] == 'ADNIGO':
                if not new_download:
                    image_dict = adnigo_image(
                            subj, visit_info[0], visits[visit_info],
                            mprage_meta_subj, ida_meta_subj, mri_quality_subj,
                            mayo_mri_qc_subj, visit_info[2])
                else:
                    image_dict = adnigo_image_refactoring(
                            csv_dir, adni_merge, subj, visit_info[0],
                            visits[visit_info], mprage_meta_subj,
                            mri_quality_subj, mayo_mri_qc_subj, visit_info[2])

            else:  # ADNI2
                image_dict = adni2_image(
                        subj, visit_info[0], visits[visit_info],
                        mprage_meta_subj_orig, mayo_mri_qc_subj)

            if image_dict is None:
                image_dict = {'Subject_ID': subj,
                              'VISCODE': visit_info[0],
                              'Visit': visits[visit_info],
                              'Sequence': '',
                              'Scan_Date': '',
                              'Study_ID': '',
                              'Series_ID': '',
                              'Image_ID': '',
                              'Field_Strength': '',
                              'Original': True}

            row_to_append = pd.DataFrame(image_dict, index=['i', ])
            t1_df = t1_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    # Subject 141_S_0726
    if not new_download:
        # conversion errors known
        subj_ind = (t1_df.Subject_ID == '141_S_0726') & (t1_df.VISCODE == 'bl')
        t1_df.loc[subj_ind, 'Sequence'] = 'MPR-R__GradWarp__B1_Correction'
        t1_df.loc[subj_ind, 'Series_ID'] = '18355'
        t1_df.loc[subj_ind, 'Image_ID'] = '94827'

        conversion_errors = [  # Eq_1
                         ('031_S_0830', 'm48'),
                         ('100_S_0995', 'm18'),
                         ('031_S_0867', 'm48'),
                         ('100_S_0892', 'm18'),
                         # Empty folders
                         ('029_S_0845', 'm24'),
                         ('094_S_1267', 'm24'),
                         ('029_S_0843', 'm24'),
                         ('027_S_0307', 'm48'),
                         ('057_S_1269', 'm24'),
                         ('036_S_4899', 'm03')]

        error_indices = []
        for conv_error in conversion_errors:
            error_indices.append((t1_df.Subject_ID == conv_error[0])
                                 & (t1_df.VISCODE == conv_error[1]))

        indices_to_remove = t1_df.index[reduce(operator.or_, error_indices, False)]
        t1_df.drop(indices_to_remove, inplace=True)

    images = t1_df
    is_dicom = []
    nifti_paths = []
    count = 0

    for row in images.iterrows():

        image = row[1]
        seq_path = path.join(source_dir, str(image.Subject_ID), image.Sequence)

        count += 1
        series_path = ''
        for (dirpath, dirnames, filenames) in walk(seq_path):
            found = False
            for d in dirnames:
                if d == 'S' + str(image.Series_ID):
                    series_path = path.join(dirpath, d)
                    found = True
                    break
            if found:
                break

        nifti_path = series_path
        dicom = True

        for (dirpath, dirnames, filenames) in walk(series_path):
            for f in filenames:
                if f.endswith(".nii"):
                    dicom = False
                    nifti_path = path.join(dirpath, f)
                    break

        is_dicom.append(dicom)
        nifti_paths.append(nifti_path)

    images.loc[:, 'Is_Dicom'] = pd.Series(is_dicom, index=images.index)
    images.loc[:, 'Path'] = pd.Series(nifti_paths, index=images.index)

    if new_download:
        images = images.drop_duplicates()
        for j in images.Path:
            if not path.exists(j):
                images.Path[images.Path == j] = ''

        images = images[images.Path != '']

    # Store the paths inside a file called conversion_info inside the input directory
    t1_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(t1_tsv_path):
        mkdir(t1_tsv_path)
    images.to_csv(path.join(t1_tsv_path, 't1_paths.tsv'), sep='\t', index=False)

    return images


def adni1_image(subject_id, timepoint, visit_str, mprage_meta_subj, ida_meta_subj, mri_quality_subj, mayo_mri_qc_subj):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint
    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                       & (mprage_meta_subj.Visit == visit_str)
                                       & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('Scaled')))]

    # If there is not a preferred image we use ADNI2 processing
    # (get the best qc if available, otherwise the original) preferring 1.5T images
    if filtered_mprage.shape[0] < 1:
        mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']
        return adni2_image(subject_id, timepoint, visit_str, mprage_meta_subj_orig, mayo_mri_qc_subj, preferred_field_strength=1.5)

    filtered_mprage_mag = filtered_mprage
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage_mag = filtered_mprage[filtered_mprage.MagStrength == 1.5]  # Select 1.5T images

    scan = filtered_mprage_mag.iloc[0]
    series_id = scan.SeriesID

    qc_passed = True
    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        # print 'QC found but NOT passed'
        # print 'Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID)
        mprage_meta_subj_alt = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Original')
                                                & (mprage_meta_subj.Visit == visit_str)
                                                & (mprage_meta_subj.SeriesID != series_id)]

        qc_prev_sequence = scan.Sequence
        scan = mprage_meta_subj_alt.iloc[0]
        series_id = scan.SeriesID
        qc_passed = False

    # TODO replace by the function to find if the scanner is_philips

    filtered_scan = ida_meta_subj[ida_meta_subj.LONIUID == series_id]

    if filtered_scan.shape[0] < 1:
        # If no IDA_META for 1.5T try for 3T
        filtered_mprage_mag = filtered_mprage[filtered_mprage.MagStrength == 3.0]
        scan = filtered_mprage_mag.iloc[0]
        series_id = scan.SeriesID

        filtered_scan = ida_meta_subj[ida_meta_subj.LONIUID == series_id]

        if filtered_scan.shape[0] < 1:
            # TODO - LOG THIS
            cprint('NO IDA Meta: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
            return None
    ida_scan = filtered_scan.iloc[0]

    original = True

    # TODO replace the condition with if is_philips
    if ida_scan.Scanner.find('Philips') > -1:

        scan = (mprage_meta_subj[
                    (mprage_meta_subj['Orig/Proc'] == 'Original') & (mprage_meta_subj.SeriesID == series_id)]).iloc[
            0]
        sequence = scan.Sequence

    else:  # scan already selected above
        sequence = scan.Sequence[:scan.Sequence.find('N3') - 2]
        original = False

    if not qc_passed:
        if scan.Sequence == 'MP-RAGE':
            original_img_seq = 'MPR'
        else:  # 'MP-RAGE REPEAT'
            original_img_seq = 'MPR-R'

        processing_seq = qc_prev_sequence[qc_prev_sequence.find(';'):qc_prev_sequence.find('N3') - 2]
        sequence = original_img_seq + processing_seq
        # print sequence

    sequence = replace_sequence_chars(sequence)

    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        # TODO - LOG THIS
        cprint('QC found but NOT passed')
        cprint('Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID))

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': original}


def adni2_image(subject_id, timepoint, visit_str, mprage_meta_subj_orig, mayo_mri_qc_subj, preferred_field_strength=3.0):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    cond_mprage = ((mprage_meta_subj_orig.Visit == visit_str)
                   & mprage_meta_subj_orig.Sequence.map(
                       lambda x: ((x.lower().find('mprage') > -1)
                                  | (x.lower().find('mp-rage') > -1)
                                  | (x.lower().find('mp rage') > -1))
                       & (x.find('2') < 0)
                       )
                   )

    cond_spgr = ((mprage_meta_subj_orig.Visit == visit_str) & mprage_meta_subj_orig.Sequence.map(
        lambda x: (x.lower().find('spgr') > -1) & (x.lower().find('acc') < 0)))

    filtered_scan = mprage_meta_subj_orig[cond_mprage | cond_spgr]

    if filtered_scan.shape[0] < 1:
        # TODO - LOG THIS
        cprint('NO MPRAGE Meta2: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
        return None

    scan = select_scan_qc_adni2(filtered_scan, mayo_mri_qc_subj, preferred_field_strength)

    sequence = replace_sequence_chars(scan.Sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': True}


def adnigo_image(subject_id, timepoint, visit_str, mprage_meta_subj, ida_meta_subj, mri_quality_subj, mayo_mri_qc_subj, original_phase):

    if original_phase == 'ADNI1':
        filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                           & (mprage_meta_subj.MagStrength == 1.5)
                                           & (mprage_meta_subj.Visit == visit_str)
                                           & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('Scaled')))]
        if filtered_mprage.shape[0] > 0:
            return adni1_image(subject_id, timepoint, visit_str, mprage_meta_subj, ida_meta_subj, mri_quality_subj, mayo_mri_qc_subj)

    mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']
    return adni2_image(subject_id, timepoint, visit_str, mprage_meta_subj_orig, mayo_mri_qc_subj)


def visits_to_timepoints_t1(subject, mprage_meta_subj_orig, adnimerge_subj):
    from datetime import datetime
    from clinica.iotools.converters.adni_to_bids.adni_utils import days_between
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj_orig[mprage_meta_subj_orig['Visit'] != 'ADNI Baseline']

    visits = dict()

    unique_visits = list(mprage_meta_subj_orig.Visit.unique())

    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():  # (adnimerge_subj[adnimerge_subj.FLDSTRENG.map(lambda x: x is not '')]).iterrows():
        visit = adni_row[1]
        if visit.ORIGPROT == 'ADNI2':
            if visit.VISCODE == 'bl':
                preferred_visit_name = 'ADNI2 Screening MRI-New Pt'
            elif visit.VISCODE == 'm03':
                preferred_visit_name = 'ADNI2 Month 3 MRI-New Pt'
            elif visit.VISCODE == 'm06':
                preferred_visit_name = 'ADNI2 Month 6-New Pt'
            else:
                year = str(int(visit.VISCODE[1:]) / 12)
                preferred_visit_name = 'ADNI2 Year ' + year + ' Visit'
        else:
            if visit.VISCODE == 'bl':
                if visit.ORIGPROT == 'ADNI1':
                    preferred_visit_name = 'ADNI Screening'
                else:  # ADNIGO
                    preferred_visit_name = 'ADNIGO Screening MRI'
            elif visit.VISCODE == 'm03':  # Only for ADNIGO Month 3
                preferred_visit_name = 'ADNIGO Month 3 MRI'
            else:
                month = int(visit.VISCODE[1:])
                if month < 54:
                    preferred_visit_name = 'ADNI1/GO Month ' + str(month)
                else:
                    preferred_visit_name = 'ADNIGO Month ' + str(month)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint('[T1] Subject ' + subject + ' has multiple visits for one timepoint. ')
                # cprint(key_preferred_visit)
                # cprint(visits[key_preferred_visit])
                # cprint(visit)
            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (mprage_meta_subj_orig[mprage_meta_subj_orig.Visit == visit]).iloc[0]
        min_db = 100000
        min_db2 = 0
        min_visit = None
        min_visit2 = None

        for timepoint in pending_timepoints:
            db = days_between(image.ScanDate, timepoint.EXAMDATE)
            if db < min_db:
                min_db2 = min_db
                min_visit2 = min_visit

                min_db = db
                min_visit = timepoint

        if min_visit is None:
            cprint('No corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit)
            cprint(image)
            continue

        if min_visit2 is not None and min_db > 90:
            cprint('More than 60 days for corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image.Visit + ' on ' + image.ScanDate)
            cprint('Timepoint 1: ' + min_visit.VISCODE + ' - ' + min_visit.ORIGPROT + ' on ' + min_visit.EXAMDATE + ' (Distance: ' + str(
                min_db) + ' days)')
            cprint('Timepoint 2: ' + min_visit2.VISCODE + ' - ' + min_visit2.ORIGPROT + ' on ' + min_visit2.EXAMDATE + ' (Distance: ' + str(
                min_db2) + ' days)')

            # If image is too close to the date between two visits we prefer the earlier visit
            if (datetime.strptime(min_visit.EXAMDATE, "%Y-%m-%d")
                    > datetime.strptime(image.ScanDate, "%Y-%m-%d")
                    > datetime.strptime(min_visit2.EXAMDATE, "%Y-%m-%d")):
                dif = days_between(min_visit.EXAMDATE, min_visit2.EXAMDATE)
                if abs((dif / 2.0) - min_db) < 30:
                    min_visit = min_visit2

            cprint('We prefer ' + min_visit.VISCODE)

        key_min_visit = (min_visit.VISCODE, min_visit.COLPROT, min_visit.ORIGPROT)
        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image.Visit
        elif visits[key_min_visit] != image.Visit:
            cprint('[T1] Subject ' + subject + ' has multiple visits for one timepoint.')
            # cprint(key_min_visit)
            # cprint(visits[key_min_visit])
            # cprint(image.Visit)
    return visits


def select_scan_no_qc(scans_meta):

    selected_scan = scans_meta[scans_meta.Sequence.map(
        lambda x: (x.lower().find('repeat') < 0))]
    if selected_scan.shape[0] < 1:
        selected_scan = scans_meta

    scan = selected_scan.iloc[0]
    return scan


def select_scan_qc_adni2(scans_meta, mayo_mri_qc_subj, preferred_field_strength):
    import numpy as np

    multiple_mag_strength = False
    # Select preferred_field_strength images
    if len(scans_meta.MagStrength.unique()) > 1:
        multiple_mag_strength = True
        not_preferred_scan = scans_meta[scans_meta.MagStrength != preferred_field_strength]
        scans_meta = scans_meta[scans_meta.MagStrength == preferred_field_strength]

    if scans_meta.MagStrength.unique()[0] == 3.0:

        id_list = scans_meta.ImageUID.unique()
        image_ids = ['I' + str(imageuid) for imageuid in id_list]
        int_ids = [int(imageuid) for imageuid in id_list]
        images_qc = mayo_mri_qc_subj[mayo_mri_qc_subj.loni_image.isin(image_ids)]

        if images_qc.shape[0] > 0:
            selected_image = None
            if np.sum(images_qc.series_selected) == 1:
                selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
            else:
                images_not_rejected = images_qc[images_qc.series_quality < 4]

                if images_not_rejected.shape[0] < 1:
                    # There are no images that passed the qc
                    # so we'll try to see if there are other images without qc,
                    # otherwise return None
                    qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
                    no_qc_ids = list(set(int_ids) - qc_ids)
                    # If none of images passed qc the scan is None, otherwise:
                    if len(no_qc_ids) > 0:
                        no_qc_scans_meta = scans_meta[scans_meta.ImageUID.isin(no_qc_ids)]
                        return select_scan_no_qc(no_qc_scans_meta)
                else:
                    series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
                    best_q = np.amin(series_quality)
                    if best_q == 4:
                        best_q = -1
                    images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
                    if images_best_qc.shape[0] == 1:
                        selected_image = images_best_qc.iloc[0].loni_image[1:]
                    else:
                        best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
                        best_qc_meta = scans_meta[scans_meta.ImageUID.isin(best_ids)]
                        return select_scan_no_qc(best_qc_meta)

            if selected_image is None and multiple_mag_strength:
                scans_meta = not_preferred_scan
            else:
                scan = scans_meta[scans_meta.ImageUID == int(selected_image)].iloc[0]
                return scan

    # 1.5T or 3.0T without QC
    return select_scan_no_qc(scans_meta)


def adni1_select_scanner(subj, csv_dir, adnimerge, timepoint):
    import pandas as pd
    import os

    com_philips = ['phil', 'n/a', 'not', 'phl', 'ND', 'na', 'n"a']
    sub_discarded = ['refused', 'declined', 'not completed', 'error', 'not on 3T scanner', '']

    mrimeta_tsv = pd.io.parsers.read_csv(os.path.join(csv_dir, 'MRIMETA.csv'), sep=',')
    mri3meta_tsv = pd.io.parsers.read_csv(os.path.join(csv_dir, 'MRI3META.csv'), sep=',')
    rid = list(adnimerge[adnimerge.PTID == subj].RID.unique())[0]

    # 1.5 T
    mrimeta_rid = mrimeta_tsv[mrimeta_tsv.RID == rid]
    mrimeta = mrimeta_rid[mrimeta_rid.VISCODE2 == timepoint]
    list_scanner_not_found = []

    if len(mrimeta) == 2:
        ix = mrimeta.index
        mrimeta = mrimeta.loc[ix[0]]

    if mrimeta.empty and timepoint == 'bl':
        timepoint = 'sc'
        mrimeta = mrimeta_rid[mrimeta_rid.VISCODE2 == timepoint]
        if len(mrimeta) > 1:

            mrimeta.drop(mrimeta.index[0], inplace=True)
        #    print mrimeta

    if not mrimeta.empty:
        scan = mrimeta.MMB1HEAD.item()
        com = mrimeta.MMB1BDCOM

        if isinstance(com, str):
            pass
        elif isinstance(com, float):
            pass
        else:
            com = com.item()

        if scan == '-4' or not scan:
            scanner = 'Philips'
        else:
            if not isinstance(com, float) and com.lower() in com_philips:
                scanner = 'Philips'
            else:
                scanner = 'GE_Siemens'
    if mrimeta.empty:
        # 3T
        mri3meta_rid = mri3meta_tsv[mri3meta_tsv.RID == rid]
        mri3meta = mri3meta_rid[mri3meta_rid.VISCODE2 == timepoint]

        if mri3meta.empty and timepoint == 'bl':
            timepoint = 'sc'
            mri3meta = mri3meta_rid[mri3meta_rid.VISCODE2 == timepoint]
        if not mri3meta.empty:
            scan = mri3meta.MMB1HEAD.item()
            com = mri3meta.MMB1BDCOM.item()
            if scan == '-4' or not scan:
                scanner = 'Philips'
            else:
                if com.lower() in com_philips:
                    scanner = 'Philips'
                else:
                    scanner = 'GE_Siemens'
        if mri3meta.empty:
            scanner = 'PROBLEM_SCANNER_NOT_FOUND'

    return scanner


def adni1_image_refactoring(csv_dir, adnimerge, subject_id, timepoint, visit_str, mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj):

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint
    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                       & (mprage_meta_subj.Visit == visit_str)
                                       & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('Scaled')))]

    # If there is not a preferred image we use ADNI2 processing
    # (get the best qc if available, otherwise the original) preferring 1.5T images
    if filtered_mprage.shape[0] < 1:
        mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']
        return adni2_image(subject_id, timepoint, visit_str, mprage_meta_subj_orig, mayo_mri_qc_subj, preferred_field_strength=1.5)

    filtered_mprage_mag = filtered_mprage
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage_mag = filtered_mprage[filtered_mprage.MagStrength == 1.5]  # Select 1.5T images

    scan = filtered_mprage_mag.iloc[0]
    series_id = scan.SeriesID

    qc_passed = True
    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        # print 'QC found but NOT passed'
        # print 'Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID)
        mprage_meta_subj_alt = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Original')
                                                & (mprage_meta_subj.Visit == visit_str)
                                                & (mprage_meta_subj.SeriesID != series_id)]

        qc_prev_sequence = scan.Sequence
        scan = mprage_meta_subj_alt.iloc[0]
        series_id = scan.SeriesID
        qc_passed = False

    # TODO replace by the function to find if the scanner is_philips

    scanner = adni1_select_scanner(subject_id, csv_dir, adnimerge, timepoint)

    original = True

    # TODO replace the condition with if is_philips
    if scanner == 'Philips':

        scan = (mprage_meta_subj[
                    (mprage_meta_subj['Orig/Proc'] == 'Original') & (mprage_meta_subj.SeriesID == series_id)]).iloc[
            0]
        sequence = scan.Sequence

    else:  # scan already selected above
        sequence = scan.Sequence[:scan.Sequence.find('N3') - 2]
        original = False

    if not qc_passed:
        if scan.Sequence == 'MP-RAGE':
            original_img_seq = 'MPR'
        else:  # 'MP-RAGE REPEAT'
            original_img_seq = 'MPR-R'

        processing_seq = qc_prev_sequence[qc_prev_sequence.find(';'):qc_prev_sequence.find('N3') - 2]
        sequence = original_img_seq + processing_seq
        # print sequence

    sequence = replace_sequence_chars(sequence)

    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]
    if qc.shape[0] > 0 and qc.iloc[0].PASS != 1:
        # TODO - LOG THIS
        cprint('QC found but NOT passed')
        cprint('Subject ' + subject_id + ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID))

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': original}


def adnigo_image_refactoring(csv_dir, adnimerge, subject_id, timepoint, visit_str, mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj, original_phase):

    if original_phase == 'ADNI1':
        filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                           & (mprage_meta_subj.MagStrength == 1.5)
                                           & (mprage_meta_subj.Visit == visit_str)
                                           & (mprage_meta_subj.Sequence.map(lambda x: x.endswith('Scaled')))]
        if filtered_mprage.shape[0] > 0:
            return adni1_image_refactoring(csv_dir, adnimerge, subject_id, timepoint, visit_str, mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj)

    mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']
    return adni2_image(subject_id, timepoint, visit_str, mprage_meta_subj_orig, mayo_mri_qc_subj)



# coding: utf-8

"""
 Module for converting FDG PET of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_fdg_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Args:
        source_dir:
        csv_dir:
        dest_dir:
        subjs_list:

    Returns:

    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.io.parsers.read_csv(adni_merge_path, sep=',')
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of FDG PET images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of FDG PET images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 'fdg')
    cprint(Fore.GREEN + 'FDG PET conversion done.' + Fore.RESET)


def compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """

    Args:
        source_dir:
        csv_dir:
        dest_dir:
        subjs_list:

    Returns:

    """
    import pandas as pd
    import os
    import operator
    from os import walk, path
    from numpy import argsort
    # from clinica.iotools.converters.adni_utils import replace_sequence_chars
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint
    from functools import reduce

    pet_fdg_col = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']

    pet_fdg_df = pd.DataFrame(columns=pet_fdg_col)
    petqc_path = path.join(csv_dir, 'PETQC.csv')
    pet_meta_list_path = path.join(csv_dir, 'PET_META_LIST.csv')
    petqc = pd.io.parsers.read_csv(petqc_path, sep=',')
    pet_meta_list = pd.io.parsers.read_csv(pet_meta_list_path, sep=',')

    for subj in subjs_list:
        pet_qc_subj = petqc[(petqc.PASS == 1) & (petqc.RID == int(subj[-4:]))]
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]
        if subject_pet_meta.shape[0] < 1:
            # TODO Log somewhere subjects with problems
            cprint('NO Screening: Subject - ' + subj)
            continue

        for visit in list(pet_qc_subj.VISCODE2.unique()):
            pet_qc_visit = pet_qc_subj[pet_qc_subj.VISCODE2 == visit]
            if pet_qc_visit.shape[0] > 1:
                normal_images = []
                normal_meta = []
                for row in pet_qc_visit.iterrows():
                    image = row[1]
                    pet_meta_image = \
                        subject_pet_meta[(subject_pet_meta['Image ID'] == int(image.LONIUID[1:]))].iloc[0]
                    if pet_meta_image.Sequence.lower().find('early') < 0:
                        normal_images.append(image)
                        normal_meta.append(pet_meta_image)
                if len(normal_images) == 0:
                    # TODO Log somewhere subjects with problems
                    cprint('No regular FDG-PET image: Subject - ' + subj + ' for visit ' + visit)
                    continue
                if len(normal_images) == 1:
                    qc_visit = normal_images[0]
                else:
                    qc_visit = None
                    index = argsort([x['Series ID'] for x in normal_meta])
                    for i in index[::-1]:
                        coreg_avg = subject_pet_meta[(subject_pet_meta['Sequence'] == 'Co-registered, Averaged')
                                                     & (
                                                         subject_pet_meta['Series ID'] == normal_meta[i][
                                                             'Series ID'])]
                        if coreg_avg.shape[0] > 0:
                            qc_visit = normal_images[i]
                            break
                    if qc_visit is None:
                        qc_visit = normal_images[index[len(index) - 1]]
            else:
                qc_visit = pet_qc_visit.iloc[0]
            int_image_id = int(qc_visit.LONIUID[1:])
            original_pet_meta = subject_pet_meta[
                (subject_pet_meta['Orig/Proc'] == 'Original') & (subject_pet_meta['Image ID'] == int_image_id)
                & (subject_pet_meta.Sequence.map(lambda s: (s.lower().find('early') < 0)))]
            if original_pet_meta.shape[0] < 1:
                original_pet_meta = subject_pet_meta[
                        (subject_pet_meta['Orig/Proc'] == 'Original')
                        & (subject_pet_meta.Sequence.map(
                            lambda x: (x.lower().find('fdg') > -1))
                           )
                        & (subject_pet_meta['Scan Date'] == qc_visit.EXAMDATE)]
                if original_pet_meta.shape[0] < 1:
                    # TODO Log somewhere subjects with problems
                    cprint('NO Screening: Subject - ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue
            original_image = original_pet_meta.iloc[0]
            averaged_pet_meta = subject_pet_meta[(subject_pet_meta['Sequence'] == 'Co-registered, Averaged') & (
                subject_pet_meta['Series ID'] == original_image['Series ID'])]
            if averaged_pet_meta.shape[0] < 1:
                sel_image = original_image
                original = True
            else:
                sel_image = averaged_pet_meta.iloc[0]
                original = False
            visit = sel_image.Visit
            sequence = replace_sequence_chars(sel_image.Sequence)
            date = sel_image['Scan Date']
            study_id = sel_image['Study ID']
            series_id = sel_image['Series ID']
            image_id = sel_image['Image ID']

            row_to_append = pd.DataFrame(
                [[subj, qc_visit.VISCODE2, str(visit), sequence, date, str(study_id), str(series_id), str(image_id),
                  original]],
                columns=pet_fdg_col)
            pet_fdg_df = pet_fdg_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # NONAME.nii
                         ('037_S_1421', 'm36'),
                         ('037_S_1078', 'm36'),
                         # Empty folders
                         ('941_S_1195', 'm48'),
                         ('005_S_0223', 'm12')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((pet_fdg_df.Subject_ID == conv_error[0])
                             & (pet_fdg_df.VISCODE == conv_error[1]))

    indices_to_remove = pet_fdg_df.index[reduce(operator.or_, error_indices, False)]
    pet_fdg_df.drop(indices_to_remove, inplace=True)

    images = pet_fdg_df
    # count = 0
    # total = images.shape[0]
    is_dicom = []
    image_folders = []
    for row in images.iterrows():
        image = row[1]
        seq_path = path.join(source_dir, str(image.Subject_ID), image.Sequence)
        # count += 1
        # print 'Processing Subject ' + str(image.Subject_ID) + ' - Session ' + image.VISCODE + ', ' + str(
        #     count) + ' / ' + str(total)
        image_path = ''
        for (dirpath, dirnames, filenames) in walk(seq_path):
            found = False
            for d in dirnames:
                if d == 'I' + str(image.Image_ID):
                    image_path = path.join(dirpath, d)
                    found = True
                    break
            if found:
                break

        dicom = True
        for (dirpath, dirnames, filenames) in walk(image_path):
            for f in filenames:
                if f.endswith(".nii"):
                    dicom = False
                    image_path = path.join(dirpath, f)
                    break

        is_dicom.append(dicom)
        image_folders.append(image_path)
        if image_path == '':
            cprint('No FDG image path found for subject ' + str(image.Subject_ID))

    images.loc[:, 'Is_Dicom'] = pd.Series(is_dicom, index=images.index)
    images.loc[:, 'Path'] = pd.Series(image_folders, index=images.index)

    fdg_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(fdg_csv_path):
        os.mkdir(fdg_csv_path)
    images.to_csv(path.join(fdg_csv_path, 'fdg_pet_paths.tsv'), sep='\t', index=False)

    return images



# coding: utf-8


def bids_dir_to_fsl_dir(bids_dir):
    """
    Converts BIDS PhaseEncodingDirection parameters (i,j,k,i-,j-,k-) to
    FSL direction (x,y,z,x-,y-,z-).
    """
    fsl_dir = bids_dir.lower()
    if fsl_dir == 'i-':
        return 'x-'
    if fsl_dir == 'i':
        return 'x'
    if fsl_dir == 'j-':
        return 'y-'
    if fsl_dir == 'j':
        return 'y'
    if fsl_dir == 'k-':
        return 'z-'
    if fsl_dir == 'k':
        return 'z'

    raise RuntimeError("PhaseEncodingDirection " + fsl_dir + " is unknown, it should be a value in (x,y,z,x-,y-,z-)")

    return fsl_dir



# coding: utf8

import clinica.pipelines.engine as cpe

__author__ = ["Thomas Jacquemont", "Alexandre Routier"]
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Nipype"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__status__ = "Development"


class DwiPreprocessingUsingT1(cpe.Pipeline):
    """DWI Preprocessing using T1 image for susceptibility distortion step.

    Warnings:
        - Do not use this pipelines if you have fieldmap data in your dataset.

    Args:
        input_dir(str): Input directory in a BIDS hierarchy.
        output_dir(str): Output directory in a CAPS hierarchy.
        subjects_sessions_list(str): The Subjects-Sessions list file (in .tsv
            format).

    Returns:
        A clinica pipeline object containing the DWIPreprocessingUsingT1 pipeline.

    Raises:


    """
    def __init__(self, bids_directory=None, caps_directory=None, tsv_file=None,
                 name=None, low_bval=5):
        """

        Args:
            bids_directory(str): Input directory in a BIDS hierarchy.
            caps_directory(str): Output directory in a CAPS hierarchy.
            tsv_file(str): TSV file containing the list of participants
                (participant_id) with their sessions (session_id).
            name(optional[str]): Name of the pipeline
            low_bval (int): Define the b0 volumes as all volume
                bval <= lowbval. (Default = 5)
        """
        import warnings

        super(DwiPreprocessingUsingT1, self).__init__(
            bids_directory=bids_directory,
            caps_directory=caps_directory,
            tsv_file=tsv_file,
            name=name)

        self._low_bval = low_bval

        if self._low_bval < 0:
            raise ValueError('The low_bval is equals to '
                             + str(self._low_bval)
                             + ': it should be zero or close to zero.')

        if self._low_bval > 100:
            warnings.warn('Warning: The low_bval parameter is ('
                          + str(self._low_bval)
                          + '), it should be close to zero', UserWarning)

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        input_list = ['T1w', 'dwi', 'bvec', 'bval']
        return input_list

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        output_list = ['preproc_dwi', 'preproc_bvec', 'preproc_bval',
                       'b0_mask']

        return output_list

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio

        from clinica.utils.stream import cprint
        from clinica.utils.dwi import check_dwi_volume

        cprint('Found %s image(s) in BIDS dataset' % len(self.subjects))
        for i in range(len(self.subjects)):
            # cprint('------- SUBJECT %s SESSION %s -------'
            #        % (self.subjects[i], self.sessions[i]))

            # Check b-val file and compute the nb of b0 from file:
            bval_file = self.bids_layout.get(
                return_type='file',
                type='dwi',
                extensions=['bval'],
                session=self.sessions[i].replace('ses-', ''),
                subject=self.subjects[i].replace('sub-', '')
            )
            if len(bval_file) != 1:
                raise IOError('Expected to find 1 bval file for subject '
                              + self.subjects[i]
                              + ' and session '
                              + self.sessions[i]
                              + ' but found '
                              + str(len(bval_file))
                              + ' bval instead.')

            # Check b-vec file:
            bvec_file = self.bids_layout.get(
                return_type='file',
                type='dwi',
                extensions=['bvec'],
                session=self.sessions[i].replace('ses-', ''),
                subject=self.subjects[i].replace('sub-', '')
            )
            if len(bvec_file) != 1:
                raise IOError('Expected to find 1 bvec file for subject '
                              + self.subjects[i]
                              + ' and session '
                              + self.sessions[i]
                              + ' but found '
                              + str(len(bvec_file))
                              + ' bvec instead.')

            # Check DWI file:
            dwi_file = self.bids_layout.get(
                return_type='file',
                type='dwi',
                extensions=['.nii|.nii.gz'],
                session=self.sessions[i].replace('ses-', ''),
                subject=self.subjects[i].replace('sub-', '')
            )
            if len(dwi_file) != 1:
                raise IOError('Expected to find 1 dwi file for subject '
                              + self.subjects[i]
                              + ' and session '
                              + self.sessions[i]
                              + ' but found '
                              + str(len(dwi_file))
                              + ' dwi instead.')

            # Check that the number of DWI, b-vecs & b-val are the same:
            check_dwi_volume(
                in_dwi=dwi_file[0], in_bvec=bvec_file[0], in_bval=bval_file[0])

            # Check T1w file:
            t1_file = self.bids_layout.get(
                return_type='file',
                type='T1w',
                extensions=['.nii|.nii.gz'],
                session=self.sessions[i].replace('ses-', ''),
                subject=self.subjects[i].replace('sub-', '')
            )
            if len(t1_file) != 1:
                raise IOError('Expected to find 1 T1w file for subject '
                              + self.subjects[i]
                              + ' and session '
                              + self.sessions[i]
                              + ' but found '
                              + str(len(t1_file))
                              + ' T1w instead.')

        # Iterables:
        iterables_node = npe.Node(name="LoadingCLIArguments",
                                  interface=nutil.IdentityInterface(
                                      fields=['subject_id', 'session_id'],
                                      mandatory_inputs=True)
                                  )
        iterables_node.iterables = [('subject_id', self.subjects),
                                    ('session_id', self.sessions)]
        iterables_node.synchronize = True

        # T1 DataGrabber
        t1_bids_reader = npe.Node(
            nio.DataGrabber(infields=['subject_id', 'session',
                                      'subject_repeat', 'session_repeat'],
                            outfields=['out_files']), name='t1_bids_reader')
        t1_bids_reader.inputs.base_directory = self.bids_directory
        t1_bids_reader.inputs.template = '%s/%s/anat/%s_%s_*T1w.nii*'
        t1_bids_reader.inputs.sort_filelist = False

        # DWI DataGrabber
        dwi_bids_reader = npe.Node(
            nio.DataGrabber(infields=['subject_id', 'session',
                                      'subject_repeat', 'session_repeat'],
                            outfields=['out_files']), name='dwi_bids_reader')
        dwi_bids_reader.inputs.base_directory = self.bids_directory
        dwi_bids_reader.inputs.template = '%s/%s/dwi/%s_%s_*dwi.nii*'
        dwi_bids_reader.inputs.sort_filelist = False

        # Bval DataGrabber
        bval_bids_reader = npe.Node(
            nio.DataGrabber(infields=['subject_id', 'session',
                                      'subject_repeat', 'session_repeat'],
                            outfields=['out_files']), name='bval_bids_reader')
        bval_bids_reader.inputs.base_directory = self.bids_directory
        bval_bids_reader.inputs.template = '%s/%s/dwi/%s_%s_*dwi.bval'
        bval_bids_reader.inputs.sort_filelist = False

        # Bvec dataGrabber
        bvec_bids_reader = npe.Node(
            nio.DataGrabber(infields=['subject_id', 'session',
                                      'subject_repeat', 'session_repeat'],
                            outfields=['out_files']), name='bvec_bids_reader')
        bvec_bids_reader.inputs.base_directory = self.bids_directory
        bvec_bids_reader.inputs.template = '%s/%s/dwi/%s_%s_*dwi.bvec'
        bvec_bids_reader.inputs.sort_filelist = False

        self.connect([
            # Iterables:
            (iterables_node,      t1_bids_reader,  [('subject_id',       'subject_id'),  # noqa
                                                    ('session_id',          'session'),  # noqa
                                                    ('subject_id',   'subject_repeat'),  # noqa
                                                    ('session_id', 'session_repeat')]),  # noqa
            (iterables_node,     dwi_bids_reader,  [('subject_id',       'subject_id'),  # noqa
                                                    ('session_id',          'session'),  # noqa
                                                    ('subject_id',   'subject_repeat'),  # noqa
                                                    ('session_id', 'session_repeat')]),  # noqa
            (iterables_node,     bval_bids_reader, [('subject_id',       'subject_id'),  # noqa
                                                    ('session_id',          'session'),  # noqa
                                                    ('subject_id',   'subject_repeat'),  # noqa
                                                    ('session_id', 'session_repeat')]),  # noqa
            (iterables_node,     bvec_bids_reader, [('subject_id',       'subject_id'),  # noqa
                                                    ('session_id',          'session'),  # noqa
                                                    ('subject_id',   'subject_repeat'),  # noqa
                                                    ('session_id', 'session_repeat')]),  # noqa
            # Inputnode:
            (t1_bids_reader,     self.input_node,  [('out_files',             'T1w')]),  # noqa
            (dwi_bids_reader,    self.input_node,  [('out_files',             'dwi')]),  # noqa
            (bval_bids_reader,   self.input_node,  [('out_files',            'bval')]),  # noqa
            (bvec_bids_reader,   self.input_node,  [('out_files',            'bvec')])   # noqa
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from os.path import join
        from clinica.utils.io import fix_join
        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils as utils

        # Find container path from DWI filename
        # =====================================
        container_path = npe.Node(nutil.Function(
            input_names=['bids_dwi_filename'],
            output_names=['container'],
            function=utils.dwi_container_from_filename),
            name='container_path')

        rename_into_caps = npe.Node(nutil.Function(
            input_names=['in_bids_dwi', 'fname_dwi', 'fname_bval',
                         'fname_bvec', 'fname_brainmask'],
            output_names=['out_caps_dwi', 'out_caps_bval', 'out_caps_bvec',
                          'out_caps_brainmask'],
            function=utils.rename_into_caps),
            name='rename_into_caps')

        # Writing results into CAPS
        # =========================
        write_results = npe.Node(name='write_results',
                                 interface=nio.DataSink())
        write_results.inputs.base_directory = self.caps_directory
        write_results.inputs.parameterization = False

        self.connect([
            (self.input_node, container_path,    [('dwi',                    'bids_dwi_filename')]),  # noqa
            (self.input_node,  rename_into_caps, [('dwi',                          'in_bids_dwi')]),  # noqa
            (self.output_node, rename_into_caps, [('preproc_dwi',                      'fname_dwi'),  # noqa
                                                  ('preproc_bval',                    'fname_bval'),  # noqa
                                                  ('preproc_bvec',                    'fname_bvec'),  # noqa
                                                  ('b0_mask',                  'fname_brainmask')]),  # noqa
            (container_path, write_results,      [(('container', fix_join, 'dwi'), 'container')]),  # noqa
            (rename_into_caps, write_results,    [('out_caps_dwi',    'preprocessing.@preproc_dwi'),  # noqa
                                                  ('out_caps_bval',  'preprocessing.@preproc_bval'),  # noqa
                                                  ('out_caps_bvec',  'preprocessing.@preproc_bvec'),  # noqa
                                                  ('out_caps_brainmask', 'preprocessing.@b0_mask')])  # noqa
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """
        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_workflows as workflows
        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils as utils

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.fsl as fsl

        from clinica.utils.dwi import prepare_reference_b0
        from clinica.workflows.dwi_preprocessing import ecc_pipeline
        from clinica.workflows.dwi_preprocessing import hmc_pipeline
        from clinica.workflows.dwi_preprocessing import remove_bias

        # Nodes creation
        # ==============

        # Prepare b0 image for further corrections
        prepare_b0 = npe.Node(name="PrepareB0", interface=nutil.Function(
            input_names=['in_dwi', 'in_bval', 'in_bvec', 'low_bval'],
            output_names=['out_reference_b0', 'out_b0_dwi_merge',
                          'out_updated_bval', 'out_updated_bvec'],
            function=prepare_reference_b0))
        prepare_b0.inputs.low_bval = self._low_bval
        # Mask b0 for computations purposes
        mask_b0_pre = npe.Node(fsl.BET(frac=0.3, mask=True, robust=True),
                               name='PreMaskB0')
        # Head-motion correction
        hmc = hmc_pipeline(name='HeadMotionCorrection')
        # Eddy-currents correction
        ecc = ecc_pipeline(name='EddyCurrentCorrection')
        # Susceptibility distortion correction using T1w image
        sdc = workflows.susceptibility_distortion_correction_using_t1(
            name='SusceptibilityDistortionCorrection')
        # Remove bias correction
        bias = remove_bias(name='RemoveBias')
        # Apply all corrections
        aac = workflows.apply_all_corrections_using_ants(
            name='ApplyAllCorrections')

        print_begin_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file'],
                function=utils.print_begin_pipeline),
            name='Write-Begin_Message')

        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file', 'final_file'],
                function=utils.print_end_pipeline),
            name='Write-End_Message')

        # Connection
        # ==========

        self.connect([
            # Print begin message
            (self.input_node, print_begin_message, [('dwi', 'in_bids_or_caps_file')]),  # noqa
            # Preliminary step (possible computation of a mean b0):
            (self.input_node, prepare_b0, [('dwi',  'in_dwi'),  # noqa
                                           ('bval', 'in_bval'),  # noqa
                                           ('bvec', 'in_bvec')]),  # noqa
            # Mask b0 before corrections
            (prepare_b0, mask_b0_pre, [('out_reference_b0', 'in_file')]),  # noqa
            # Head-motion correction
            (prepare_b0,  hmc, [('out_b0_dwi_merge', 'inputnode.in_file'),  # noqa
                               ('out_updated_bval',  'inputnode.in_bval'),  # noqa
                               ('out_updated_bvec',  'inputnode.in_bvec')]),  # noqa
            (mask_b0_pre, hmc, [('mask_file',        'inputnode.in_mask')]),  # noqa
            # Eddy-current correction
            (hmc,         ecc, [('outputnode.out_xfms', 'inputnode.in_xfms')]),  # noqa
            (prepare_b0,  ecc, [('out_b0_dwi_merge',    'inputnode.in_file')]),  # noqa
            (prepare_b0,  ecc, [('out_updated_bval',    'inputnode.in_bval')]),  # noqa
            (mask_b0_pre, ecc, [('mask_file',           'inputnode.in_mask')]),  # noqa
            # Magnetic susceptibility correction
            (self.input_node, sdc, [('T1w',                 'inputnode.in_t1')]),  # noqa
            (ecc,             sdc, [('outputnode.out_file', 'inputnode.in_dwi')]),  # noqa
            (hmc,             sdc, [('outputnode.out_bvec', 'inputnode.in_bvec')]),  # noqa
            # Apply all corrections
            (prepare_b0,      aac, [('out_b0_dwi_merge',    'inputnode.in_dwi')]),  # noqa
            (hmc,             aac, [('outputnode.out_xfms', 'inputnode.in_hmc')]),  # noqa
            (ecc,             aac, [('outputnode.out_xfms', 'inputnode.in_ecc')]),  # noqa
            (sdc,             aac, [('outputnode.out_warp', 'inputnode.in_sdc_syb')]),  # noqa
            (self.input_node, aac, [('T1w',                 'inputnode.in_t1')]),  # noqa
            # Bias correction
            (aac, bias, [('outputnode.out_file', 'inputnode.in_file')]),
            # Outputnode:
            (bias,       self.output_node, [('outputnode.out_file', 'preproc_dwi')]),  # noqa
            (sdc,        self.output_node, [('outputnode.out_bvec', 'preproc_bvec')]),  # noqa
            (prepare_b0, self.output_node, [('out_updated_bval',    'preproc_bval')]),  # noqa
            (bias,       self.output_node, [('outputnode.b0_mask',  'b0_mask')]),  # noqa
            # Print end message
            (self.input_node, print_end_message, [('dwi',         'in_bids_or_caps_file')]),  # noqa
            (bias,            print_end_message, [('outputnode.out_file', 'final_file')]),  # noqa
        ])



# coding: utf8


def erode_mask(in_mask, npass=6, nthreads=2):
    """
    Erode the mask.

    This function is used for the erosion of the brain mask thus preventing
    voxels near the edge of the brain from potentially being erroneously
    selected as single-fibre voxels during the estimation of the response
    function.

    Args:
        in_mask (str): Binary mask.
        npass (Optional[int]): Number of times to repeatedly apply the erosion
            (default=6 which is used in MRtrix community).
        nthreads (Optional[int]): Number of threads used in this function
            (default=2, 1 disables multi-threading).

    Returns:
        An eroded mask.
    """
    import os.path as op
    import os

    assert(op.isfile(in_mask))

    out_eroded_mask = op.abspath('eroded_mask.nii.gz')
    cmd = 'maskfilter -npass %s -nthreads %s %s erode %s' \
          % (npass, nthreads, in_mask, out_eroded_mask)
    os.system(cmd)
    return out_eroded_mask


def statistics_on_atlases(in_registered_map, name_map, prefix_file=None):
    """
    Computes a list of statistics files for each atlas.

    Args:
        in_registered_map (str): Map already registered on atlases.
        name_map (str): Name of the registered map in CAPS format.
        prefix_file (Opt[str]):
            <prefix_file>_space-<atlas_name>_map-<name_map>_statistics.tsv

    Returns:
        List of paths leading to the statistics TSV files.
    """
    from os import getcwd
    from os.path import abspath, join
    from nipype.utils.filemanip import split_filename
    from clinica.utils.atlas import (AtlasAbstract, JHUDTI811mm,
                                     JHUTracts01mm, JHUTracts251mm)
    from clinica.utils.statistics import statistics_on_atlas

    in_atlas_list = [JHUDTI811mm(),
                     JHUTracts01mm(), JHUTracts251mm()]

    atlas_statistics_list = []
    for atlas in in_atlas_list:
        if not isinstance(atlas, AtlasAbstract):
            raise TypeError("Atlas element must be an AtlasAbstract type")

        if prefix_file is None:
            _, base, _ = split_filename(in_registered_map)
            filename = '%s_space-%s_res-%s_map-%s_statistics.tsv' % \
                       (base, atlas.get_name_atlas(),
                        atlas.get_spatial_resolution(), name_map)
        else:
            filename = '%s_space-%s_res-%s_map-%s_statistics.tsv' % \
                       (prefix_file, atlas.get_name_atlas(),
                        atlas.get_spatial_resolution(), name_map)

        out_atlas_statistics = abspath(join(getcwd(), filename))

        statistics_on_atlas(in_registered_map, atlas, out_atlas_statistics)
        atlas_statistics_list.append(out_atlas_statistics)

    return atlas_statistics_list


def dwi_container_from_filename(dwi_filename):
    import re
    from os.path import join
    m = re.search(r'(sub-[a-zA-Z0-9]+)/(ses-[a-zA-Z0-9]+)', dwi_filename)

    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.'
            ' It does not contain the subject and session information.')

    subject = m.group(1)
    session = m.group(2)

    return join('subjects', subject, session)


def extract_bids_identifier_from_caps_filename(caps_dwi_filename):
    """Extract BIDS identifier from CAPS filename"""
    import re

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi',
                  caps_dwi_filename)

    if m is None:
        raise ValueError('Input filename is not in a CAPS compliant format.')
    bids_identifier = m.group(0)

    return bids_identifier


def get_caps_filenames(caps_dwi_filename):
    """
    Prepare some filenames with CAPS naming convention
    """
    import re

    m = re.search(
        r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi_space-[a-zA-Z0-9]+',
        caps_dwi_filename)
    if m is None:
        raise ValueError('Input filename is not in a CAPS compliant format.')

    caps_prefix = m.group(0)
    bids_source = m.group(1) + '_' + m.group(2) + '_dwi'

    out_dti = caps_prefix + '_model-DTI_diffmodel.nii.gz'
    out_fa = caps_prefix + '_FA.nii.gz'
    out_md = caps_prefix + '_MD.nii.gz'
    out_ad = caps_prefix + '_AD.nii.gz'
    out_rd = caps_prefix + '_RD.nii.gz'

    return bids_source, out_dti, out_fa, out_md, out_ad, out_rd


def rename_into_caps(in_caps_dwi,
                     in_norm_fa, in_norm_md, in_norm_ad, in_norm_rd,
                     in_b_spline_transform, in_affine_matrix):
    """
    Rename different outputs of the pipelines into CAPS format.

    Returns:
        The different outputs with CAPS naming convention
    """
    from nipype.interfaces.utility import Rename
    from clinica.pipelines.dwi_dti.dwi_dti_utils import extract_bids_identifier_from_caps_filename

    bids_identifier = extract_bids_identifier_from_caps_filename(in_caps_dwi)

    # CAPS normalized FA
    rename_fa = Rename()
    rename_fa.inputs.in_file = in_norm_fa
    rename_fa.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_FA.nii.gz"
    out_caps_fa = rename_fa.run()
    # CAPS normalized MD
    rename_md = Rename()
    rename_md.inputs.in_file = in_norm_md
    rename_md.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_MD.nii.gz"
    out_caps_md = rename_md.run()
    # CAPS normalized AD
    rename_ad = Rename()
    rename_ad.inputs.in_file = in_norm_ad
    rename_ad.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_AD.nii.gz"
    out_caps_ad = rename_ad.run()
    # CAPS normalized RD
    rename_rd = Rename()
    rename_rd.inputs.in_file = in_norm_rd
    rename_rd.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_RD.nii.gz"
    out_caps_rd = rename_rd.run()
    # CAPS B-spline transform
    rename_b_spline = Rename()
    rename_b_spline.inputs.in_file = in_b_spline_transform
    rename_b_spline.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_deformation.nii.gz"
    out_caps_b_spline_transform = rename_b_spline.run()
    # CAPS Affine matrix
    rename_affine = Rename()
    rename_affine.inputs.in_file = in_affine_matrix
    rename_affine.inputs.format_string = bids_identifier + "_space-MNI152Lin_res-1x1x1_affine.mat"
    out_caps_affine_matrix = rename_affine.run()

    return out_caps_fa.outputs.out_file, out_caps_md.outputs.out_file,\
        out_caps_ad.outputs.out_file, out_caps_rd.outputs.out_file, \
        out_caps_b_spline_transform.outputs.out_file, \
        out_caps_affine_matrix.outputs.out_file


def print_begin_pipeline(in_bids_or_caps_file):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s Running pipeline for %s...' % (
        Fore.BLUE, now, Fore.RESET, m.group(0)))


def print_end_pipeline(in_bids_or_caps_file, final_file_1, final_file_2):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s ...%s has completed.' % (
        Fore.GREEN, now, Fore.RESET, m.group(0)))



# coding: utf8

"""
Methods to find information in the different pipelines of Clinica
"""
__author__ = "Elina Thubeau--Sutre"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = [""]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Simona Bottani"
__email__ = "simona.bottani@icm-institute.org"
__status__ = "Completed"

from glob import glob
import pandas as pd
import numpy as np
from os import path
import os
import warnings


def pet_volume_pipeline(caps_dir, df, **kwargs):
    """
    This method merge the data of the PET-Volume pipeline to the merged file
    containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        atlas_selection: allows to choose the atlas to merge (default = 'all')
        pvc_restriction: gives the restriction on the inclusion or not of the file with the label 'pvc-rbv'
            1       --> only the atlases containing the label will be used
            0       --> the atlases containing the label won't be used
            None    --> all the atlases will be used

    Returns:
         final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from clinica.pipelines.pet_volume.pet_volume_cli import PETVolumeCLI
    instance = PETVolumeCLI()
    pipeline_name = instance.name
    del instance

    # Recovering optional arguments
    if 'pvc_restriction' in kwargs.keys():
        pvc_restriction = kwargs['pvc_restriction']
    else:
        pvc_restriction = None

    if 'atlas_selection' in kwargs.keys():
        atlas_selection = kwargs['atlas_selection']
    else:
        atlas_selection = None

    if 'group_selection' in kwargs.keys():
        group_selection = kwargs['group_selection']
    else:
        group_selection = None

    if pvc_restriction not in [None, 0, 1]:
        raise ValueError('Error in pvc_restriction value: ' + str(pvc_restriction))

    if not path.isdir(caps_dir):
        raise IOError("The CAPS directory doesn't exist")

    pet_path = 'pet' + os.sep + 'preprocessing'
    caps_dir = caps_dir + os.sep + 'subjects'

    # Lists to be sure of the column in case of missing data
    columns = ['pipeline_name', 'group_id', 'atlas_id', 'regions_number', 'first_column_name', 'last_column_name']
    summary_df = pd.DataFrame(columns=columns)
    col_list = []
    group_list = []

    # Looking for a subject with the correct modality to initialize
    sub_paths = glob(path.join(caps_dir, '*sub-*'))
    flag_initial_subject = True
    i = 0
    while flag_initial_subject:
        sub_path = sub_paths[i]
        i += 1
        ses_paths = glob(path.join(sub_path, '*ses-*'))
        for ses_path in ses_paths:
            possible_mods = os.listdir(ses_path)
            if 'pet' in possible_mods:
                flag_initial_subject = False
                ses_init_path = ses_path
            elif i == len(sub_paths):
                raise InitException(pipeline_name)

    # Initialization must take into account the different groups !
    init_path = ses_init_path + os.sep + pet_path
    if group_selection is None:
        group_paths = glob(path.join(init_path, '*group*'))
    else:
        group_paths = []
        for group_selected in group_selection:
            new_group_path = glob(path.join(init_path, '*' + group_selected + '*'))
            group_paths = group_paths + new_group_path
            if len(new_group_path) == 0:
                warnings.warn('The group wanted does not exist for ' + pipeline_name + ': ' + group_selected, UserWarning)
    for group_path in group_paths:
        group_id = group_path.split(os.sep)[-1]
        group_list.append(group_id)
        stats_path = group_path + os.sep + 'atlas_statistics'
        # Selection of the atlas(es)
        if atlas_selection is None:
            atlas_paths = glob(path.join(stats_path, '*statistics*'))
        else:
            atlas_paths = []
            for atlas_selected in atlas_selection:
                new_atlas_path = glob(path.join(stats_path, '*' + atlas_selected + '*'))
                atlas_paths = atlas_paths + new_atlas_path
                if len(new_atlas_path) == 0:
                    warnings.warn('The atlas wanted does not exist for ' + pipeline_name + ': ' + atlas_selected + ' in group ' + group_id, UserWarning)

        if len(atlas_paths) == 0:
                raise InitException(pipeline_name)

        atlas_paths_preserved = list(atlas_paths)
        # Selection depending on the label 'pvc-rbv'
        for i, atlas_path in enumerate(atlas_paths):
            atlas_split = atlas_path.split(os.sep)[-1].split('_')
            if pvc_restriction == 1 and 'pvc-rbv' != atlas_split[6]:
                atlas_paths_preserved.remove(atlas_path)
            elif pvc_restriction == 0 and 'pvc-rbv' == atlas_split[6]:
                atlas_paths_preserved.remove(atlas_path)

        atlas_paths = atlas_paths_preserved
        atlas_paths.sort()
        for atlas_path in atlas_paths:
            row_summary_df = pd.DataFrame(index=np.arange(1), columns=columns)
            atlas_file = atlas_path.split(os.sep)[-1]
            atlas_id = '_'.join(atlas_file.split('_')[2:-1])  # We get rid of the subject and the session in the label

            init_df = pd.read_csv(atlas_path, sep='\t')
            n_regions = len(init_df['label_name'].tolist())
            label_list = [group_id + '_' + atlas_id + '_ROI-' + str(x) for x in range(n_regions)]
            for label in label_list:
                col_list.append(label)

            row_summary = [pipeline_name, group_id, atlas_id, n_regions, label_list[0], label_list[-1]]
            row_summary_df.iloc[0] = row_summary
            summary_df = pd.concat([summary_df, row_summary_df])

    # Initial empty DataFrame for the pipeline
    number_sessions = len(df)
    pipeline_df = pd.DataFrame(index=np.arange(number_sessions), columns=col_list)

    for i in range(number_sessions):
        # Reading the reference (subject + session)
        participant_id = df['participant_id'][i]
        session_id = df['session_id'][i]

        ses_path = caps_dir + os.sep + participant_id + os.sep + session_id
        mod_path = ses_path + os.sep + pet_path
        ses_df = pd.DataFrame(index=np.arange(1), columns=col_list)

        if os.path.exists(mod_path):
            # Looking for groups
            for group in group_list:
                group_path = mod_path + os.sep + group
                if os.path.exists(group_path):
                    # Looking for atlases
                    atlas_list = summary_df[summary_df.group_id == group]['atlas_id'].values
                    for atlas in atlas_list:
                        atlas_path = group_path + os.sep + 'atlas_statistics' + os.sep
                        atlas_path = atlas_path + participant_id + '_' + session_id + '_' + atlas + '_statistics.tsv'
                        if os.path.exists(atlas_path):
                            n_regions = summary_df[(summary_df.group_id == group) &
                                                   (summary_df.atlas_id == atlas)]['regions_number'].values[0]
                            atlas_df = pd.read_csv(atlas_path, sep='\t')
                            label_list = [group + '_' + atlas + '_ROI-' + str(x) for x in
                                          range(n_regions)]
                            ses_df.loc[:, label_list] = atlas_df['mean_scalar'].as_matrix()

        pipeline_df.iloc[i] = ses_df.iloc[0]

    final_df = pd.concat([df, pipeline_df], axis=1)

    return final_df, summary_df


def t1_volume_pipeline(caps_dir, df, **kwargs):
    """
    This method merge the data of the t1-volume pipeline to the merged file containing the
    BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        atlas_selection: allows to choose the atlas to merge. If None all atlases are selected.

    Returns:
        final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from ...pipelines.t1_volume_new_template.t1_volume_new_template_cli import T1VolumeNewTemplateCLI
    instance = T1VolumeNewTemplateCLI()
    pipeline_name = instance.name
    del instance

    # Recovering optional arguments
    if 'atlas_selection' in kwargs.keys():
        atlas_selection = kwargs['atlas_selection']
    else:
        atlas_selection = None

    if 'group_selection' in kwargs.keys():
        group_selection = kwargs['group_selection']
    else:
        group_selection = None

    t1_spm_path = 't1' + os.sep + 'spm' + os.sep + 'dartel'
    caps_dir = caps_dir + os.sep + 'subjects'

    # Lists to be sure of the column in case of missing data
    columns = ['pipeline_name', 'group_id', 'atlas_id', 'regions_number', 'first_column_name', 'last_column_name']
    summary_df = pd.DataFrame(columns=columns)
    col_list = []
    group_list = []

    # Looking for a subject with the correct modality to initialize
    sub_paths = glob(path.join(caps_dir, '*sub-*'))
    flag_initial_subject = True
    i = 0
    while flag_initial_subject:
        sub_path = sub_paths[i]
        i += 1
        ses_paths = glob(path.join(sub_path, '*ses-*'))
        for ses_path in ses_paths:
            possible_mods = os.listdir(ses_path)
            if 't1' in possible_mods:
                mod_path = path.join(ses_path, 't1')
                possible_mods = os.listdir(mod_path)
                if 'spm' in possible_mods:
                    flag_initial_subject = False
                    ses_init_path = ses_path
            elif i == len(sub_paths):
                raise InitException(pipeline_name)

    # Initialization must take into account the different groups !
    init_path = ses_init_path + os.sep + t1_spm_path
    if group_selection is None:
        group_paths = glob(path.join(init_path, '*group*'))
    else:
        group_paths = []
        for group_selected in group_selection:
            new_group_path = glob(path.join(init_path, '*' + group_selected + '*'))
            group_paths = group_paths + new_group_path
            if len(new_group_path) == 0:
                warnings.warn('The group wanted does not exist for ' + pipeline_name + ': ' + group_selected, UserWarning)
    for group_path in group_paths:
        group_id = group_path.split(os.sep)[-1]
        group_list.append(group_id)
        stats_path = group_path + os.sep + 'atlas_statistics'
        # Selection of the atlas(es)
        if atlas_selection is None:
            atlas_paths = glob(path.join(stats_path, '*statistics*'))
        else:
            atlas_paths = []
            for atlas_selected in atlas_selection:
                new_atlas_path = glob(path.join(stats_path, '*' + atlas_selected + '*'))
                atlas_paths = atlas_paths + new_atlas_path
                if len(new_atlas_path) == 0:
                    warnings.warn('The atlas wanted does not exist for ' + pipeline_name + ': ' + atlas_selected + ' in group ' + group_id, UserWarning)

        if len(atlas_paths) == 0:
                raise InitException(pipeline_name)

        atlas_paths.sort()
        for atlas_path in atlas_paths:
            row_summary_df = pd.DataFrame(index=np.arange(1), columns=columns)
            atlas_file = atlas_path.split(os.sep)[-1]
            atlas_id = '_'.join(atlas_file.split('_')[2:-1])  # We get rid of the subject and the session in the label

            init_df = pd.read_csv(atlas_path, sep='\t')
            n_regions = len(init_df['label_name'].tolist())
            label_list = [group_id + '_' + atlas_id + '_ROI-' + str(x) for x in range(n_regions)]
            for label in label_list:
                col_list.append(label)

            row_summary = [pipeline_name, group_id, atlas_id, n_regions, label_list[0], label_list[-1]]
            row_summary_df.iloc[0] = row_summary
            summary_df = pd.concat([summary_df, row_summary_df])

    # Initial empty DataFrame for the pipeline
    number_sessions = len(df)
    pipeline_df = pd.DataFrame(index=np.arange(number_sessions), columns=col_list)

    for i in range(number_sessions):
        # Reading the reference (subject + session)
        participant_id = df['participant_id'][i]
        session_id = df['session_id'][i]

        ses_path = caps_dir + os.sep + participant_id + os.sep + session_id
        mod_path = ses_path + os.sep + t1_spm_path
        ses_df = pd.DataFrame(index=np.arange(1), columns=col_list)

        if os.path.exists(mod_path):
            # Looking for groups
            for group in group_list:
                group_path = mod_path + os.sep + group
                if os.path.exists(group_path):
                    # Looking for atlases
                    atlas_list = summary_df[summary_df.group_id == group]['atlas_id'].values
                    for atlas in atlas_list:
                        atlas_path = group_path + os.sep + 'atlas_statistics' + os.sep
                        atlas_path = atlas_path + participant_id + '_' + session_id + '_' + atlas + '_statistics.tsv'
                        if os.path.exists(atlas_path):
                            n_regions = summary_df[(summary_df.group_id == group) &
                                                   (summary_df.atlas_id == atlas)]['regions_number'].values[0]
                            atlas_df = pd.read_csv(atlas_path, sep='\t')
                            label_list = [group + '_' + atlas + '_ROI-' + str(x) for x in
                                          range(n_regions)]
                            ses_df.loc[:, label_list] = atlas_df['mean_scalar'].as_matrix()

        pipeline_df.iloc[i] = ses_df.iloc[0]

    final_df = pd.concat([df, pipeline_df], axis=1)

    return final_df, summary_df


class InitException(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr('The following pipeline cannot be initialized: ' + self.name)


class DatasetError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr('Bad format for the sessions: ' + self.name)



# coding: utf8


import numpy as np
from sklearn.metrics import balanced_accuracy_score, accuracy_score, classification_report


def evaluate_prediction(y, y_hat):

    true_positive = 0.0
    true_negative = 0.0
    false_positive = 0.0
    false_negative = 0.0

    tp = []
    tn = []
    fp = []
    fn = []

    for i in range(len(y)):
        if y[i] == 1:
            if y_hat[i] == 1:
                true_positive += 1
                tp.append(i)
            else:
                false_negative += 1
                fn.append(i)
        else:  # -1
            if y_hat[i] == 0:
                true_negative += 1
                tn.append(i)
            else:
                false_positive += 1
                fp.append(i)

    accuracy = (true_positive + true_negative) / (true_positive + true_negative + false_positive + false_negative)

    if (true_positive + false_negative) != 0:
        sensitivity = true_positive / (true_positive + false_negative)
    else:
        sensitivity = 0.0

    if (false_positive + true_negative) != 0:
        specificity = true_negative / (false_positive + true_negative)
    else:
        specificity = 0.0

    if (true_positive + false_positive) != 0:
        ppv = true_positive / (true_positive + false_positive)
    else:
        ppv = 0.0

    if (true_negative + false_negative) != 0:
        npv = true_negative / (true_negative + false_negative)
    else:
        npv = 0.0

    balanced_accuracy = (sensitivity + specificity) / 2

    results = {'accuracy': accuracy,
               'balanced_accuracy': balanced_accuracy,
               'sensitivity': sensitivity,
               'specificity': specificity,
               'ppv': ppv,
               'npv': npv,
               'confusion_matrix': {'tp': len(tp), 'tn': len(tn), 'fp': len(fp), 'fn': len(fn)}
               }

    return results


def gram_matrix_linear(data):
    return np.dot(data, data.transpose())


def evaluate_prediction_multiclass(y, y_hat):

    balanced_accuracy = balanced_accuracy_score(y, y_hat)
    accuracy = accuracy_score(y, y_hat)

    results = {'accuracy': accuracy,
               'balanced_accuracy': balanced_accuracy}

    return results



# coding: utf8

import clinica.pipelines.engine as cpe

__author__ = "Simona Bottani"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Simona Bottani"
__email__ = "simona.bottani@icm-institute.org"
__status__ = "Development"


class T1VolumeParcellation(cpe.Pipeline):
    """T1VolumeParcellation - Computation of mean GM concentration for a set of regions

    Args: input_dir: A BIDS directory.  output_dir: An empty output directory
    where CAPS structured data will be written.  subjects_sessions_list: The
    Subjects-Sessions list file (in .tsv format).

    Returns:
        A clinica pipeline object containing the T1VolumeParcellation pipeline.

    Raises:

    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['file_list', 'atlas_list']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        pass

    def build_input_node(self):
        """Build and connect an input node to the pipeline.
        """

        from clinica.utils.stream import cprint
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.lib.pycaps.caps_layout import CAPSLayout

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.atlas_list = self.parameters['atlases']

        caps_layout = CAPSLayout(self.caps_directory)

        cprint('------- INPUT FILES FOR EACH SUBJECTS -------')

        subjects_regex = '|'.join(sub[4:] for sub in self.subjects)
        unique_session = set(list(self.sessions))
        sessions_regex = '|'.join(sub[4:] for sub in unique_session)
        cprint('  * grabbing all files from CAPS folder')
        caps_file = caps_layout.get(return_type='file',
                                    subject=subjects_regex,
                                    session=sessions_regex,
                                    group_id=self.parameters['group_id'],
                                    modulation=self.parameters['modulate'])
        if len(caps_file) != len(self.subjects):
            raise IOError(str(len(caps_file)) + ' file(s) grabbed, but there is ' + str(len(self.subjects)) + ' sessions')
        read_parameters_node.inputs.file_list = caps_file

        self.connect([(read_parameters_node, self.input_node, [('file_list', 'file_list')]),
                      (read_parameters_node, self.input_node, [('atlas_list', 'atlas_list')])
                      ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline.
        """

        # In the same idea as the input node, this output node is supposedly
        # used to write the output fields in a CAPS. It should be executed only
        # if this pipeline output is not already connected to a next Clinica
        # pipeline.
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import clinica.pipelines.t1_volume_parcellation.t1_volume_parcellation_utils as spu
        import nipype.interfaces.io as nio

        atlas_stats_node = npe.MapNode(nutil.Function(input_names=['file_list',
                                                                   'atlas_list'],
                                                      output_names=['atlas_statistics'],
                                                      function=spu.atlas_statistics),
                                       name='atlas_stats_node',
                                       iterfield=['file_list'])
        outputnode = npe.Node(nutil.IdentityInterface(fields=['atlas_statistics']),
                              name='outputnode',
                              mandatory_inputs=True)

        datasink = npe.Node(nio.DataSink(),
                            name='datasink')

        datasink.inputs.base_directory = self.caps_directory
        datasink.inputs.parameterization = True
        datasink.inputs.regexp_substitutions = [
            (r'(.*)(atlas_statistics)/.*/(sub-(.*)_ses-(.*)_T1.*)$',
             r'\1/subjects/sub-\4/ses-\5/t1/spm/dartel/group-' + self.parameters['group_id'] + r'/\2/\3')]

        # Connection
        # ==========
        self.connect([
            (self.input_node,      atlas_stats_node,    [('file_list',    'file_list')]),
            (self.input_node,      atlas_stats_node,    [('atlas_list',    'atlas_list')]),
            (atlas_stats_node,     outputnode,          [('atlas_statistics',  'atlas_statistics')]),
            (outputnode,           datasink,            [('atlas_statistics', 'atlas_statistics')])
        ])



# coding: utf8


def convert_phase_in_radians(in_file, out_file=None):
    """
    Convert phase image in radians.
    """
    import math
    import nibabel as nb
    import numpy as np
    import os.path as op
    import os

    assert(op.isfile(in_file))

    img = nb.load(in_file)
    imin = np.amin(img.get_data())
    imax = np.amax(img.get_data())

    if out_file is None:
        out_file = op.abspath('phase_in_rad.nii.gz')

    cmd = 'fslmaths %s -mul %s -div %s %s -odt float' \
          % (in_file, 2.0*math.pi, imax, out_file)
    os.system(cmd)

    data = (img2.get_data().astype(np.float32)-img1.get_data().astype(np.float32)) * (1.0/delta_te)
    nb.Nifti1Image(data, img.get_affine(), img.get_header()).to_filename(out_file)

    return out_file


def create_phase_in_radsec(in_phase1, in_phase2, delta_te, out_file=None):
    """
    Converts input (unwarpped) phase1 and phase2 map to into a fieldmap in
    rads. (delta_te should be in seconds)
    """
    import numpy as np
    import nibabel as nb
    import os.path as op

    if out_file is None:
        out_file = op.abspath('fmap_radsec.nii.gz')

    img1 = nb.load(in_phase1)
    img2 = nb.load(in_phase2)
    data = (img2.get_data().astype(np.float32)-img1.get_data().astype(np.float32)) * (1.0/delta_te)
    nb.Nifti1Image(data, img1.get_affine(), img1.get_header()).to_filename(out_file)
    return out_file


def resample_fmap_to_b0(in_fmap, in_b0, out_file=None):
    """
    Resample fieldmap onto the b0 image.

    Warnings:
        The fieldmap should already be aligned on the b0.

    Args:
        in_fmap(str): Fieldmap image.
        in_b0(str): B0 image.
        out_file(optional[str]): Filename (default: <in_fmap>_space-b0.nii.gz

    Returns:
        Fieldmap image resampled on b0.
    """
    import os.path as op
    import nibabel
    from nilearn.image import resample_to_img

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_fmap))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_resampled_fmap = op.abspath("%s_space-b0%s" % (fname, ext))
    else:
        out_resampled_fmap = out_file

    resampled_fmap = resample_to_img(
        source_img=in_fmap, target_img=in_b0,
        interpolation='continuous')

    nibabel.nifti1.save(resampled_fmap, out_resampled_fmap)

    return out_resampled_fmap
